from __future__ import annotations

import shutil
from dataclasses import dataclass, field
from pathlib import Path

from evals import run_eval
from evals.adaptive_online_policy import (
    ACTION_INVOKE_HELPER,
    ACTION_SKIP_HELPER,
    STATE_SCHEMA_VERSION,
    AdaptiveOnlinePolicyController,
    OnlinePolicyConfig,
)
from evals.patchers.base import ContextRequestResult, PatchResult
from evals.patchers.openai_compat_patcher import OpenAICompatPatcher
from llmdebug import _json as json


def _make_case(tmp_path: Path) -> run_eval.CaseSpec:
    case_dir = tmp_path / "case_src"
    case_dir.mkdir(parents=True, exist_ok=True)
    (case_dir / "buggy.py").write_text(
        "import helper\n\n\ndef f() -> int:\n    return helper.value()\n",
        encoding="utf-8",
    )
    (case_dir / "helper.py").write_text("def value() -> int:\n    return 1\n", encoding="utf-8")
    (case_dir / "test_buggy.py").write_text(
        "from buggy import f\n\n\ndef test_f() -> None:\n    assert f() == 2\n",
        encoding="utf-8",
    )
    return run_eval.CaseSpec(
        case_id="adaptive_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="helpful",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )


def _snapshot_payload() -> dict:
    return {
        "exception": {"type": "AssertionError", "message": "expected 2 got 1"},
        "frames": [
            {
                "file_rel": "buggy.py",
                "line": 4,
                "function": "f",
                "code": "return helper.value()",
                "locals": {"x": 1},
            }
        ],
        "pytest": {"nodeid": "test_buggy.py::test_f", "repro": ["python", "-m", "pytest", "-q"]},
    }


def _skip_biased_online_controller(
    tmp_path: Path, *, mode: str
) -> tuple[AdaptiveOnlinePolicyController, Path]:
    state_path = tmp_path / f"policy_state_{mode}.json"
    events_path = tmp_path / f"policy_events_{mode}.jsonl"
    state_path.write_text(
        json.dumps(
            {
                "schema_version": STATE_SCHEMA_VERSION,
                "mode": mode,
                "algorithm": "linucb",
                "alpha": 0.0,
                "epsilon": 0.0,
                "l2": 1.0,
                "warmup_decisions": 0,
                "save_every": 1,
                "feature_names": ["bias"],
                "decisions": 0,
                "updates": 0,
                "action_counts": {
                    ACTION_SKIP_HELPER: 0,
                    ACTION_INVOKE_HELPER: 0,
                },
                "A": {
                    ACTION_SKIP_HELPER: [[1.0]],
                    ACTION_INVOKE_HELPER: [[1.0]],
                },
                "b": {
                    ACTION_SKIP_HELPER: [2.0],
                    ACTION_INVOKE_HELPER: [0.0],
                },
            },
            ensure_ascii=False,
        ),
        encoding="utf-8",
    )
    config = OnlinePolicyConfig.normalize(
        mode=mode,
        algorithm="linucb",
        alpha=0.0,
        epsilon=0.0,
        l2=1.0,
        warmup_decisions=0,
        save_every=1,
        state_path=state_path,
        events_path=events_path,
        random_seed=0,
    )
    return AdaptiveOnlinePolicyController(config), events_path


@dataclass
class _AdaptiveGatedCapturePatcher:
    name: str = "adaptive_gated_capture"
    request_calls: int = 0
    prompts: list[str] = field(default_factory=list)

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt, context_dir, cwd
        self.request_calls += 1
        return ContextRequestResult(
            request_json={
                "need_more_context": True,
                "requests": [{"type": "file", "path": "buggy.py"}],
                "sufficiency_rationale": "need code context",
            },
            raw_response='{"need_more_context":true}',
            meta={},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del context_dir, cwd
        self.prompts.append(prompt)
        return PatchResult(diff=None, meta={"reason": "no_diff"})


@dataclass
class _AdaptiveDiscretionDeclinePatcher:
    name: str = "adaptive_discretion_decline"
    request_calls: int = 0

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt, context_dir, cwd
        self.request_calls += 1
        return ContextRequestResult(
            request_json={
                "need_more_context": False,
                "requests": [],
                "sufficiency_rationale": "enough evidence",
            },
            raw_response='{"need_more_context":false,"requests":[]}',
            meta={},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        return PatchResult(diff=None, meta={"reason": "no_diff"})


@dataclass
class _AdaptiveDiscretionInvalidPatcher:
    name: str = "adaptive_discretion_invalid"
    request_calls: int = 0
    prompts: list[str] = field(default_factory=list)

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt, context_dir, cwd
        self.request_calls += 1
        return ContextRequestResult(request_json=None, raw_response="{invalid-json", meta={})

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del context_dir, cwd
        self.prompts.append(prompt)
        return PatchResult(diff=None, meta={"reason": "no_diff"})


@dataclass
class _AdaptiveDiscretionSnapshotGenericPatcher:
    name: str = "adaptive_discretion_snapshot_generic"
    request_calls: int = 0
    prompts: list[str] = field(default_factory=list)

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt, context_dir, cwd
        self.request_calls += 1
        return ContextRequestResult(
            request_json={
                "need_more_context": False,
                "need_snapshot": True,
                "snapshot_justification": "Need more context.",
                "requests": [],
                "sufficiency_rationale": "need state",
            },
            raw_response='{"need_more_context":false,"need_snapshot":true}',
            meta={},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del context_dir, cwd
        self.prompts.append(prompt)
        return PatchResult(diff=None, meta={"reason": "no_diff"})


@dataclass
class _AdaptiveToggleDiffPatcher:
    name: str = "adaptive_toggle_diff"
    request_calls: int = 0
    patch_calls: int = 0

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt, context_dir, cwd
        self.request_calls += 1
        return ContextRequestResult(
            request_json={
                "need_more_context": True,
                "requests": [{"type": "file", "path": "buggy.py"}],
                "sufficiency_rationale": "need source",
            },
            raw_response='{"need_more_context":true}',
            meta={},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        self.patch_calls += 1
        if self.patch_calls % 2 == 1:
            diff = (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def value() -> int:\n"
                "-    return 1\n"
                "+    return 2\n"
            )
        else:
            diff = (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def value() -> int:\n"
                "-    return 2\n"
                "+    return 1\n"
            )
        return PatchResult(diff=diff, meta={"reason": "toggle_diff"})


@dataclass
class _AdaptiveTestEditPatcher:
    name: str = "adaptive_test_edit"
    request_calls: int = 0

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt, context_dir, cwd
        self.request_calls += 1
        return ContextRequestResult(
            request_json={
                "need_more_context": True,
                "requests": [{"type": "file", "path": "buggy.py"}],
                "sufficiency_rationale": "need source",
            },
            raw_response='{"need_more_context":true}',
            meta={},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        diff = (
            "diff --git a/test_buggy.py b/test_buggy.py\n"
            "--- a/test_buggy.py\n"
            "+++ b/test_buggy.py\n"
            "@@ -3,4 +3,4 @@\n"
            " \n"
            " \n"
            " def test_f() -> None:\n"
            "-    assert f() == 2\n"
            "+    assert f() == 1\n"
        )
        return PatchResult(diff=diff, meta={"reason": "test_edit"})


@dataclass
class _AdaptiveNeighborEditPatcher:
    name: str = "adaptive_neighbor_edit"
    request_calls: int = 0

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del prompt, context_dir, cwd
        self.request_calls += 1
        return ContextRequestResult(
            request_json={
                "need_more_context": True,
                "requests": [{"type": "file", "path": "buggy.py"}],
                "sufficiency_rationale": "need source",
            },
            raw_response='{"need_more_context":true}',
            meta={},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        diff = (
            "diff --git a/helper.py b/helper.py\n"
            "--- a/helper.py\n"
            "+++ b/helper.py\n"
            "@@ -1,2 +1,2 @@\n"
            " def value() -> int:\n"
            "-    return 1\n"
            "+    return 2\n"
        )
        return PatchResult(diff=diff, meta={"reason": "neighbor_edit"})


@dataclass
class _SyntaxCyclingPatcher:
    name: str = "syntax_cycling"
    diffs: list[str] = field(default_factory=list)
    metas: list[dict[str, object]] = field(default_factory=list)
    patch_calls: int = 0

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        self.patch_calls += 1
        if not self.diffs:
            return PatchResult(diff=None, meta={"reason": "no_diff"})
        idx = min(self.patch_calls - 1, len(self.diffs) - 1)
        meta: dict[str, object] = {"reason": "cycle_diff", "llm_call_count": 1}
        if self.metas:
            meta.update(self.metas[min(idx, len(self.metas) - 1)])
        return PatchResult(diff=self.diffs[idx], meta=meta)


def test_adaptive_gated_invokes_helper_after_repeated_signature(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_gated"
    context_dir = tmp_path / "context_adaptive_gated"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveGatedCapturePatcher()

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_GATED,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=2,
        output_format="json_compact",
        run_id="run-adaptive-gated",
        include_files_context=False,
        force_patch_attempt=False,
        adaptive_gate_policy="balanced",
        adaptive_stagnation_window=1,
    )

    assert len(records) == 2
    assert patcher.request_calls == 1
    first_attempt_meta = records[0]["attempt_metadata"]
    second_attempt_meta = records[1]["attempt_metadata"]
    assert first_attempt_meta["adaptive_helper_invoked"] is False
    assert second_attempt_meta["adaptive_helper_invoked"] is True
    assert first_attempt_meta["context_protocol"] == "monolithic"
    assert second_attempt_meta["context_protocol"] == "staged"
    assert records[1]["patch_meta"]["adaptive_helper_decision_reason"] in {
        "gate_balanced_repeated_signature",
        "gate_requested_helper",
    }
    assert len(patcher.prompts) == 2
    assert "Pytest output (stdout):" in patcher.prompts[1]
    assert "Stage-A evidence index:" not in patcher.prompts[1]
    assert second_attempt_meta["retry_delta_prompt"] is False


def test_adaptive_llm_discretion_decline_keeps_monolithic_attempt(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_discretion_decline"
    context_dir = tmp_path / "context_adaptive_discretion_decline"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveDiscretionDeclinePatcher()

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-adaptive-discretion-decline",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    assert patcher.request_calls == 1
    attempt_meta = records[0]["attempt_metadata"]
    assert attempt_meta["adaptive_helper_invoked"] is False
    assert attempt_meta["adaptive_helper_decision_reason"] == "llm_declined_helper"
    assert attempt_meta["context_protocol"] == "monolithic"
    assert records[0]["patch_meta"]["stage_a_request_success"] is True


def test_adaptive_openai_helper_response_mode_override(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_openai_override"
    context_dir = tmp_path / "context_adaptive_openai_override"
    shutil.copytree(case.path, workdir)

    captured_response_modes: list[str] = []

    def _fake_request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del self, prompt, context_dir, cwd
        return ContextRequestResult(
            request_json={
                "need_more_context": True,
                "requests": [{"type": "file", "path": "buggy.py"}],
                "sufficiency_rationale": "need source",
            },
            raw_response='{"need_more_context":true}',
            meta={},
        )

    def _fake_propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        captured_response_modes.append(str(self.response_mode))
        return PatchResult(
            diff=None, meta={"reason": "no_diff", "response_mode": self.response_mode}
        )

    monkeypatch.setattr(OpenAICompatPatcher, "request_context", _fake_request_context)
    monkeypatch.setattr(OpenAICompatPatcher, "propose_patch", _fake_propose_patch)
    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    patcher = OpenAICompatPatcher(
        base_url="http://127.0.0.1:1234",
        model="dummy",
        response_mode="edits",
        force_patch_attempt=False,
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-adaptive-openai-override",
        include_files_context=False,
        force_patch_attempt=False,
        adaptive_openai_helper_response_mode="diff",
    )

    assert len(records) == 1
    assert captured_response_modes == ["diff"]
    assert records[0]["patch_meta"]["adaptive_openai_response_mode_override"] == "diff"
    assert records[0]["model_metadata"]["response_mode"] == "diff"


def test_adaptive_llm_discretion_invalid_response_does_not_count_as_invocation(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_discretion_invalid"
    context_dir = tmp_path / "context_adaptive_discretion_invalid"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveDiscretionInvalidPatcher()

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-adaptive-discretion-invalid",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    assert patcher.request_calls == 1
    assert len(patcher.prompts) == 1
    assert "Stage-A requested context:" not in patcher.prompts[0]
    attempt_meta = records[0]["attempt_metadata"]
    assert attempt_meta["adaptive_helper_invoked"] is False
    assert attempt_meta["adaptive_helper_decision_reason"] == "llm_invalid_response_no_helper"
    assert attempt_meta["context_protocol"] == "monolithic"
    assert records[0]["patch_meta"]["stage_a_request_success"] is False


def test_adaptive_llm_discretion_attempt1_snapshot_denial_syncs_helper_reason(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_discretion_snapshot_denial"
    context_dir = tmp_path / "context_adaptive_discretion_snapshot_denial"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveDiscretionSnapshotGenericPatcher()

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-adaptive-discretion-snapshot-denial",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    assert patcher.request_calls == 1
    attempt_meta = records[0]["attempt_metadata"]
    patch_meta = records[0]["patch_meta"]
    assert attempt_meta["adaptive_snapshot_attempt1_denied"] is True
    assert attempt_meta["adaptive_snapshot_included"] is False
    assert attempt_meta["adaptive_helper_invoked"] is False
    assert attempt_meta["adaptive_helper_decision_reason"] == "llm_snapshot_request_denied_attempt1"
    assert patch_meta["adaptive_helper_decision_reason"] == "llm_snapshot_request_denied_attempt1"
    assert attempt_meta["context_protocol"] == "monolithic"


def test_adaptive_llm_discretion_caps_helper_calls_and_stops_on_no_progress(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_discretion_cap"
    context_dir = tmp_path / "context_adaptive_discretion_cap"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveGatedCapturePatcher(name="adaptive_discretion_cap")

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=5,
        output_format="json_compact",
        run_id="run-adaptive-discretion-cap",
        include_files_context=False,
        force_patch_attempt=False,
        adaptive_max_helper_calls=1,
        adaptive_budget_topup_max_calls=0,
    )

    assert len(records) == 3
    assert patcher.request_calls == 1
    assert len(patcher.prompts) == 3
    assert records[1]["attempt_metadata"]["adaptive_helper_budget_exhausted"] is True
    assert records[2]["attempt_metadata"]["adaptive_progress_stop_triggered"] is True
    assert records[2]["patch_meta"]["adaptive_progress_stop_reason"] == "helper_budget_exhausted"
    assert records[2]["note"] == "no_patch_proposed"


def test_adaptive_llm_discretion_grants_one_budget_topup(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_discretion_topup"
    context_dir = tmp_path / "context_adaptive_discretion_topup"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveGatedCapturePatcher(name="adaptive_discretion_topup")

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=4,
        output_format="json_compact",
        run_id="run-adaptive-discretion-topup",
        include_files_context=False,
        force_patch_attempt=False,
        adaptive_max_helper_calls=1,
        adaptive_budget_topup_max_calls=1,
    )

    assert len(records) == 4
    assert patcher.request_calls == 2
    topup_attempts = [
        record
        for record in records
        if bool(record["attempt_metadata"]["adaptive_budget_topup_granted"])
    ]
    assert len(topup_attempts) == 1
    assert (
        topup_attempts[0]["attempt_metadata"]["adaptive_budget_topup_reason"]
        == "need_more_context_recovery_once"
    )
    assert records[-1]["attempt_metadata"]["adaptive_helper_budget_exhausted"] is True
    assert records[-1]["attempt_metadata"]["adaptive_progress_stop_triggered"] is True


def test_snapshot_justification_generic_detector_handles_variants() -> None:
    assert run_eval._is_generic_snapshot_justification("Need more context.")
    assert run_eval._is_generic_snapshot_justification("Need the snapshot, please!")
    assert not run_eval._is_generic_snapshot_justification(
        "Need snapshot to inspect locals in frame 0 for None-path divergence."
    )


def test_adaptive_reintroduces_stage_a_index_when_context_is_truncated(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_index_fallback"
    context_dir = tmp_path / "context_adaptive_index_fallback"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveGatedCapturePatcher(name="adaptive_index_fallback")

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-adaptive-index-fallback",
        include_files_context=False,
        force_patch_attempt=False,
        prompt_budget_total_chars=900,
        prompt_budget_evidence_chars=80,
    )

    assert len(records) == 1
    assert patcher.request_calls == 1
    assert len(patcher.prompts) == 1
    assert records[0]["attempt_metadata"]["stage_a_index_included"] is True
    assert records[0]["attempt_metadata"]["stage_a_index_reason"] == "context_truncated_fallback"


def test_adaptive_postpatch_recompute_avoids_premature_stop_on_signature_change(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_postpatch_recompute"
    context_dir = tmp_path / "context_adaptive_postpatch_recompute"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveToggleDiffPatcher()

    run_outputs = [
        "sig_a",  # initial run
        "sig_a",  # after attempt 1
        "sig_a",  # after attempt 2
        "sig_b",  # after attempt 3 (changed)
        "sig_b",  # after attempt 4
    ]
    call_counter = {"idx": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        idx = call_counter["idx"]
        call_counter["idx"] += 1
        signature = run_outputs[min(idx, len(run_outputs) - 1)]
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout=f"{signature}\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )
    monkeypatch.setattr(
        run_eval,
        "_build_failure_signature_for_attempt",
        lambda *, run, snapshot: (run.stdout or "").strip(),
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=4,
        output_format="json_compact",
        run_id="run-adaptive-postpatch-recompute",
        include_files_context=False,
        force_patch_attempt=False,
        adaptive_max_helper_calls=1,
    )

    assert len(records) == 4
    third_attempt = records[2]
    assert third_attempt["attempt"] == 3
    assert third_attempt["patch_meta"]["adaptive_progress_recomputed_after_patch"] is True
    assert third_attempt["attempt_metadata"]["adaptive_progress_stop_triggered"] is False


def test_adaptive_patch_verifier_rejects_test_only_patch_before_apply(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_verifier_reject"
    context_dir = tmp_path / "context_adaptive_verifier_reject"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveTestEditPatcher()

    run_calls = {"count": 0}
    apply_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    def _fake_apply_diff(cwd, diff):
        del cwd, diff
        apply_calls["count"] += 1
        return True, ""

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(run_eval, "apply_diff", _fake_apply_diff)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-adaptive-verifier-reject",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    assert run_calls["count"] == 1
    assert apply_calls["count"] == 0
    assert records[0]["note"] == "patch_verifier_rejected"
    assert records[0]["patch_meta"]["patch_verifier_enabled"] is True
    assert records[0]["patch_meta"]["patch_verifier_scope"] == "adaptive_only"
    assert records[0]["patch_meta"]["patch_verifier_passed"] is False
    assert records[0]["attempt_metadata"]["patch_verifier_passed"] is False


def test_adaptive_patch_verifier_allows_neighbor_patch_and_reaches_apply(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_verifier_pass"
    context_dir = tmp_path / "context_adaptive_verifier_pass"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveNeighborEditPatcher()

    run_calls = {"count": 0}
    apply_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    def _fake_apply_diff(cwd, diff):
        del cwd, diff
        apply_calls["count"] += 1
        return False, "simulated apply failure"

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(run_eval, "apply_diff", _fake_apply_diff)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-adaptive-verifier-pass",
        include_files_context=False,
        force_patch_attempt=False,
    )

    assert len(records) == 1
    assert run_calls["count"] == 1
    assert apply_calls["count"] == 1
    assert records[0]["note"] == "patch_apply_failed"
    assert records[0]["patch_meta"]["patch_verifier_enabled"] is True
    assert records[0]["patch_meta"]["patch_verifier_passed"] is True


def test_adaptive_patch_verifier_infers_source_scope_from_failing_test_file(tmp_path) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_verifier_scope_infer"
    shutil.copytree(case.path, workdir)

    diff = (
        "diff --git a/buggy.py b/buggy.py\n"
        "--- a/buggy.py\n"
        "+++ b/buggy.py\n"
        "@@ -2,4 +2,4 @@\n"
        " \n"
        " \n"
        " def f() -> int:\n"
        "-    return helper.value()\n"
        "+    return helper.value() + 0\n"
    )
    snapshot = {
        "exception": {"type": "AssertionError", "message": "expected 2 got 1"},
        "frames": [
            {
                "file_rel": "tests/test_buggy.py",
                "line": 4,
                "function": "test_f",
                "code": "assert f() == 2",
                "locals": {},
            }
        ],
    }
    stdout = "test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n"

    passed_without_infer, meta_without_infer = run_eval._score_patch_relevance(
        diff=diff,
        cwd=workdir,
        snapshot=snapshot,
        stdout=stdout,
        stderr="",
        infer_test_source_scope=False,
    )
    passed_with_infer, meta_with_infer = run_eval._score_patch_relevance(
        diff=diff,
        cwd=workdir,
        snapshot=snapshot,
        stdout=stdout,
        stderr="",
        infer_test_source_scope=True,
    )

    assert passed_without_infer is False
    assert meta_without_infer["patch_verifier_source_scope_matched"] is False
    assert passed_with_infer is True
    assert meta_with_infer["patch_verifier_source_scope_matched"] is True
    assert "buggy.py" in meta_with_infer["patch_verifier_inferred_source_scope"]


def test_adaptive_patch_verifier_infers_scope_from_subprocess_target(tmp_path) -> None:
    workdir = tmp_path / "workdir_adaptive_verifier_subprocess_scope"
    (workdir / "tests").mkdir(parents=True, exist_ok=True)
    (workdir / "buggy.py").write_text("def f() -> int:\n    return 1\n", encoding="utf-8")
    (workdir / "tests" / "test_cli.py").write_text(
        (
            "import subprocess\n\n"
            "def test_cli() -> None:\n"
            "    subprocess.run(['python', 'buggy.py'])\n"
            "    assert True\n"
        ),
        encoding="utf-8",
    )

    diff = (
        "diff --git a/buggy.py b/buggy.py\n"
        "--- a/buggy.py\n"
        "+++ b/buggy.py\n"
        "@@ -1,2 +1,2 @@\n"
        " def f() -> int:\n"
        "-    return 1\n"
        "+    return 2\n"
    )
    snapshot = {
        "exception": {"type": "AssertionError", "message": "cli mismatch"},
        "frames": [
            {
                "file_rel": "tests/test_cli.py",
                "line": 4,
                "function": "test_cli",
                "code": "assert True",
                "locals": {},
            }
        ],
    }
    stdout = "tests/test_cli.py::test_cli\nE   AssertionError: cli mismatch\n"

    passed_without_infer, _meta_without_infer = run_eval._score_patch_relevance(
        diff=diff,
        cwd=workdir,
        snapshot=snapshot,
        stdout=stdout,
        stderr="",
        infer_test_source_scope=False,
    )
    passed_with_infer, meta_with_infer = run_eval._score_patch_relevance(
        diff=diff,
        cwd=workdir,
        snapshot=snapshot,
        stdout=stdout,
        stderr="",
        infer_test_source_scope=True,
    )

    assert passed_without_infer is False
    assert passed_with_infer is True
    assert meta_with_infer["patch_verifier_source_scope_matched"] is True
    assert "buggy.py" in meta_with_infer["patch_verifier_inferred_source_scope"]


def test_patch_verifier_scope_match_avoids_ambiguous_basename_false_positive() -> None:
    ambiguous_scope = {"pkg/a/utils.py", "pkg/b/utils.py"}
    assert (
        run_eval._patch_mentions_scope_files(
            patch_files=["other/utils.py"],
            scope_files=ambiguous_scope,
        )
        is False
    )
    assert (
        run_eval._patch_mentions_scope_files(
            patch_files=["utils.py"],
            scope_files=ambiguous_scope,
        )
        is False
    )

    unambiguous_scope = {"pkg/a/buggy.py"}
    assert (
        run_eval._patch_mentions_scope_files(
            patch_files=["buggy.py"],
            scope_files=unambiguous_scope,
        )
        is True
    )
    assert (
        run_eval._patch_mentions_scope_files(
            patch_files=["other/buggy.py"],
            scope_files=unambiguous_scope,
        )
        is False
    )


def test_syntax_preflight_retries_within_same_attempt(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_syntax_preflight_retry"
    context_dir = tmp_path / "context_syntax_preflight_retry"
    shutil.copytree(case.path, workdir)
    patcher = _SyntaxCyclingPatcher(
        name="syntax_preflight_retry",
        diffs=[
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def value() -> int:\n"
                "-    return 1\n"
                "+    return (\n"
            ),
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def value() -> int:\n"
                "-    return 1\n"
                "+    return 2\n"
            ),
        ],
    )

    run_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        if run_calls["count"] == 1:
            return run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
                stderr="",
            )
        helper_text = (workdir / "helper.py").read_text(encoding="utf-8")
        if "return 2" in helper_text and "return (" not in helper_text:
            return run_eval.RunResult(
                returncode=0,
                duration_sec=0.01,
                stdout="1 passed\n",
                stderr="",
            )
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   SyntaxError: '(' was never closed\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.WITH_SNAPSHOT,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-syntax-preflight-retry",
        include_files_context=False,
        force_patch_attempt=False,
        syntax_preflight_enabled=True,
        syntax_preflight_max_cycles=2,
    )

    assert len(records) == 1
    assert run_calls["count"] == 2
    assert patcher.patch_calls == 2
    assert records[0]["success"] is True
    assert records[0]["note"] == "success"
    assert records[0]["attempt"] == 1
    assert records[0]["attempt_metadata"]["syntax_preflight_cycles_used"] == 2
    assert records[0]["attempt_metadata"]["syntax_preflight_failures"] == 1
    events = records[0]["patch_meta"]["syntax_preflight_events"]
    assert len(events) == 2
    assert events[0]["passed"] is False
    assert events[1]["passed"] is True


def test_syntax_preflight_exhaustion_marks_attempt_failed(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_syntax_preflight_failed"
    context_dir = tmp_path / "context_syntax_preflight_failed"
    shutil.copytree(case.path, workdir)
    patcher = _SyntaxCyclingPatcher(
        name="syntax_preflight_failed",
        diffs=[
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def value() -> int:\n"
                "-    return 1\n"
                "+    return (\n"
            )
        ],
    )

    run_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.WITH_SNAPSHOT,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-syntax-preflight-failed",
        include_files_context=False,
        force_patch_attempt=False,
        syntax_preflight_enabled=True,
        syntax_preflight_max_cycles=2,
    )

    assert len(records) == 1
    assert run_calls["count"] == 1
    assert patcher.patch_calls == 2
    assert records[0]["success"] is False
    assert records[0]["note"] == "syntax_preflight_failed"
    assert records[0]["attempt_metadata"]["syntax_preflight_cycles_used"] == 2
    assert records[0]["attempt_metadata"]["syntax_preflight_failures"] == 2
    assert records[0]["attempt_metadata"]["syntax_preflight_last_error_type"] == "PyCompileError"
    events = records[0]["patch_meta"]["syntax_preflight_events"]
    assert len(events) == 2
    assert all(event["passed"] is False for event in events)


def test_syntax_preflight_clears_transient_patch_meta_between_cycles(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_syntax_preflight_meta_cleanup"
    context_dir = tmp_path / "context_syntax_preflight_meta_cleanup"
    shutil.copytree(case.path, workdir)
    patcher = _SyntaxCyclingPatcher(
        name="syntax_preflight_meta_cleanup",
        diffs=[
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def value() -> int:\n"
                "-    return 1\n"
                "+    return (\n"
            ),
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def value() -> int:\n"
                "-    return 1\n"
                "+    return 2\n"
            ),
        ],
        metas=[
            {"runner_forced_patch_attempt": True},
            {},
        ],
    )

    run_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        if run_calls["count"] == 1:
            return run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
                stderr="",
            )
        return run_eval.RunResult(
            returncode=0,
            duration_sec=0.01,
            stdout="1 passed\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.WITH_SNAPSHOT,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-syntax-preflight-meta-cleanup",
        include_files_context=False,
        force_patch_attempt=False,
        syntax_preflight_enabled=True,
        syntax_preflight_max_cycles=2,
    )

    assert len(records) == 1
    assert records[0]["success"] is True
    assert patcher.patch_calls == 2
    assert "runner_forced_patch_attempt" not in records[0]["patch_meta"]


def test_patch_sanitize_reject_then_retry_success(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_patch_sanitize_retry"
    context_dir = tmp_path / "context_patch_sanitize_retry"
    shutil.copytree(case.path, workdir)
    patcher = _SyntaxCyclingPatcher(
        name="patch_sanitize_retry",
        diffs=[
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,3 @@\n"
                " def value() -> int:\n"
                "+    # keep \\n marker\n"
                "     return 1\n"
            ),
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,2 @@\n"
                " def value() -> int:\n"
                "-    return 1\n"
                "+    return 2\n"
            ),
        ],
    )

    run_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        if run_calls["count"] == 1:
            return run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
                stderr="",
            )
        return run_eval.RunResult(
            returncode=0,
            duration_sec=0.01,
            stdout="1 passed\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.WITH_SNAPSHOT,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-patch-sanitize-retry",
        include_files_context=False,
        force_patch_attempt=False,
        syntax_preflight_enabled=True,
        syntax_preflight_max_cycles=2,
        patch_sanitize_enabled=True,
    )

    assert len(records) == 1
    assert records[0]["success"] is True
    assert run_calls["count"] == 2
    assert patcher.patch_calls == 2
    assert records[0]["attempt_metadata"]["syntax_preflight_failures"] == 1
    events = records[0]["patch_meta"]["syntax_preflight_events"]
    assert len(events) == 2
    assert events[0]["passed"] is False
    assert events[0]["syntax_preflight_error_type"] == "PatchSanitizationError"
    assert events[1]["passed"] is True


def test_patch_sanitize_exhaustion_marks_attempt_failed(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_patch_sanitize_failed"
    context_dir = tmp_path / "context_patch_sanitize_failed"
    shutil.copytree(case.path, workdir)
    patcher = _SyntaxCyclingPatcher(
        name="patch_sanitize_failed",
        diffs=[
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,3 @@\n"
                " def value() -> int:\n"
                "+    # keep \\n marker\n"
                "     return 1\n"
            )
        ],
    )

    run_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.WITH_SNAPSHOT,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-patch-sanitize-failed",
        include_files_context=False,
        force_patch_attempt=False,
        syntax_preflight_enabled=True,
        syntax_preflight_max_cycles=2,
        patch_sanitize_enabled=True,
    )

    assert len(records) == 1
    assert records[0]["success"] is False
    assert records[0]["note"] == "syntax_preflight_failed"
    assert run_calls["count"] == 1
    assert patcher.patch_calls == 2
    assert records[0]["attempt_metadata"]["syntax_preflight_failures"] == 2
    assert (
        records[0]["attempt_metadata"]["syntax_preflight_last_error_type"]
        == "PatchSanitizationError"
    )
    assert records[0]["patch_meta"]["patch_sanitize_rejected"] is True
    assert records[0]["patch_meta"]["patch_sanitize_reasons"]


def test_patch_sanitize_disabled_allows_patch_to_apply(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_patch_sanitize_disabled"
    context_dir = tmp_path / "context_patch_sanitize_disabled"
    shutil.copytree(case.path, workdir)
    patcher = _SyntaxCyclingPatcher(
        name="patch_sanitize_disabled",
        diffs=[
            (
                "diff --git a/helper.py b/helper.py\n"
                "--- a/helper.py\n"
                "+++ b/helper.py\n"
                "@@ -1,2 +1,3 @@\n"
                " def value() -> int:\n"
                "+    # keep \\n marker\n"
                "     return 1\n"
            )
        ],
    )

    run_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        if run_calls["count"] == 1:
            return run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
                stderr="",
            )
        return run_eval.RunResult(
            returncode=0,
            duration_sec=0.01,
            stdout="1 passed\n",
            stderr="",
        )

    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.WITH_SNAPSHOT,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-patch-sanitize-disabled",
        include_files_context=False,
        force_patch_attempt=False,
        syntax_preflight_enabled=False,
        patch_sanitize_enabled=False,
    )

    assert len(records) == 1
    assert records[0]["success"] is True
    assert run_calls["count"] == 2
    assert patcher.patch_calls == 1
    assert records[0]["patch_meta"]["patch_sanitize_enabled"] is False
    assert records[0]["patch_meta"]["patch_sanitize_rejected"] is False


def test_openai_response_mode_autoswitch_on_patch_sanitize_retry(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_openai_autoswitch_sanitize"
    context_dir = tmp_path / "context_openai_autoswitch_sanitize"
    shutil.copytree(case.path, workdir)

    captured_response_modes: list[str] = []
    diffs = [
        (
            "diff --git a/helper.py b/helper.py\n"
            "--- a/helper.py\n"
            "+++ b/helper.py\n"
            "@@ -1,2 +1,3 @@\n"
            " def value() -> int:\n"
            "+    # keep \\n marker\n"
            "     return 1\n"
        ),
        (
            "diff --git a/helper.py b/helper.py\n"
            "--- a/helper.py\n"
            "+++ b/helper.py\n"
            "@@ -1,2 +1,2 @@\n"
            " def value() -> int:\n"
            "-    return 1\n"
            "+    return 2\n"
        ),
    ]

    def _fake_propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        captured_response_modes.append(str(self.response_mode))
        idx = min(len(captured_response_modes) - 1, len(diffs) - 1)
        return PatchResult(diff=diffs[idx], meta={"response_mode": self.response_mode})

    run_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        if run_calls["count"] == 1:
            return run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
                stderr="",
            )
        return run_eval.RunResult(
            returncode=0,
            duration_sec=0.01,
            stdout="1 passed\n",
            stderr="",
        )

    monkeypatch.setattr(OpenAICompatPatcher, "propose_patch", _fake_propose_patch)
    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    patcher = OpenAICompatPatcher(
        base_url="http://127.0.0.1:1234",
        model="dummy",
        response_mode="edits",
        force_patch_attempt=False,
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.WITH_SNAPSHOT,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-openai-autoswitch-sanitize",
        include_files_context=False,
        force_patch_attempt=False,
        syntax_preflight_enabled=True,
        syntax_preflight_max_cycles=2,
        patch_sanitize_enabled=True,
        openai_response_mode_autoswitch="retry_to_diff",
    )

    assert len(records) == 1
    assert records[0]["success"] is True
    assert run_calls["count"] == 2
    assert captured_response_modes == ["edits", "diff"]
    assert records[0]["patch_meta"]["openai_response_mode_autoswitched"] is True
    assert records[0]["patch_meta"]["openai_response_mode_autoswitch_reason"] == (
        "patch_sanitize_rejected"
    )
    assert records[0]["model_metadata"]["response_mode"] == "diff"


def test_openai_response_mode_autoswitch_on_syntax_preflight_retry(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_openai_autoswitch_syntax"
    context_dir = tmp_path / "context_openai_autoswitch_syntax"
    shutil.copytree(case.path, workdir)

    captured_response_modes: list[str] = []
    diffs = [
        (
            "diff --git a/helper.py b/helper.py\n"
            "--- a/helper.py\n"
            "+++ b/helper.py\n"
            "@@ -1,2 +1,2 @@\n"
            " def value() -> int:\n"
            "-    return 1\n"
            "+    return (\n"
        ),
        (
            "diff --git a/helper.py b/helper.py\n"
            "--- a/helper.py\n"
            "+++ b/helper.py\n"
            "@@ -1,2 +1,2 @@\n"
            " def value() -> int:\n"
            "-    return 1\n"
            "+    return 2\n"
        ),
    ]

    def _fake_propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del prompt, context_dir, cwd
        captured_response_modes.append(str(self.response_mode))
        idx = min(len(captured_response_modes) - 1, len(diffs) - 1)
        return PatchResult(diff=diffs[idx], meta={"response_mode": self.response_mode})

    run_calls = {"count": 0}

    def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
        del args, kwargs
        run_calls["count"] += 1
        if run_calls["count"] == 1:
            return run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
                stderr="",
            )
        helper_text = (workdir / "helper.py").read_text(encoding="utf-8")
        if "return 2" in helper_text and "return (" not in helper_text:
            return run_eval.RunResult(
                returncode=0,
                duration_sec=0.01,
                stdout="1 passed\n",
                stderr="",
            )
        return run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="E   SyntaxError: '(' was never closed\n",
            stderr="",
        )

    monkeypatch.setattr(OpenAICompatPatcher, "propose_patch", _fake_propose_patch)
    monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    patcher = OpenAICompatPatcher(
        base_url="http://127.0.0.1:1234",
        model="dummy",
        response_mode="edits",
        force_patch_attempt=False,
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.WITH_SNAPSHOT,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=1,
        output_format="json_compact",
        run_id="run-openai-autoswitch-syntax",
        include_files_context=False,
        force_patch_attempt=False,
        syntax_preflight_enabled=True,
        syntax_preflight_max_cycles=2,
        patch_sanitize_enabled=False,
        openai_response_mode_autoswitch="retry_to_diff",
    )

    assert len(records) == 1
    assert records[0]["success"] is True
    assert run_calls["count"] == 2
    assert captured_response_modes == ["edits", "diff"]
    assert records[0]["patch_meta"]["openai_response_mode_autoswitched"] is True
    assert records[0]["patch_meta"]["openai_response_mode_autoswitch_reason"] == (
        "syntax_preflight_failed"
    )
    assert records[0]["model_metadata"]["response_mode"] == "diff"


def test_adaptive_online_policy_shadow_keeps_baseline_action(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_online_shadow"
    context_dir = tmp_path / "context_adaptive_online_shadow"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveGatedCapturePatcher(name="adaptive_online_shadow")
    controller, events_path = _skip_biased_online_controller(tmp_path, mode="shadow")

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    try:
        records = run_eval.run_one_condition(
            case=case,
            condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
            workdir=workdir,
            context_dir_base=context_dir,
            patcher=patcher,
            k=1,
            output_format="json_compact",
            run_id="run-adaptive-online-shadow",
            include_files_context=False,
            force_patch_attempt=False,
            adaptive_online_policy_controller=controller,
        )
    finally:
        controller.close()

    assert len(records) == 1
    assert patcher.request_calls == 1
    attempt_meta = records[0]["attempt_metadata"]
    assert attempt_meta["adaptive_online_policy_mode"] == "shadow"
    assert attempt_meta["adaptive_online_action_selected"] == ACTION_SKIP_HELPER
    assert attempt_meta["adaptive_online_action_applied"] == ACTION_INVOKE_HELPER
    assert attempt_meta["adaptive_online_override_reason"] == "shadow_mode_baseline"
    assert attempt_meta["adaptive_online_updated"] is False
    assert attempt_meta["adaptive_online_update_reason"] == "shadow_no_update"
    assert attempt_meta["adaptive_online_event_logged"] is True
    events = [line for line in events_path.read_text(encoding="utf-8").splitlines() if line.strip()]
    assert len(events) == 1


def test_adaptive_online_policy_on_can_skip_helper(tmp_path, monkeypatch) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_online_on"
    context_dir = tmp_path / "context_adaptive_online_on"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveGatedCapturePatcher(name="adaptive_online_on")
    controller, events_path = _skip_biased_online_controller(tmp_path, mode="on")

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    try:
        records = run_eval.run_one_condition(
            case=case,
            condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
            workdir=workdir,
            context_dir_base=context_dir,
            patcher=patcher,
            k=1,
            output_format="json_compact",
            run_id="run-adaptive-online-on",
            include_files_context=False,
            force_patch_attempt=False,
            adaptive_online_policy_controller=controller,
        )
    finally:
        controller.close()

    assert len(records) == 1
    assert patcher.request_calls == 0
    attempt_meta = records[0]["attempt_metadata"]
    assert attempt_meta["adaptive_helper_decision_reason"] == "online_policy_skip_helper"
    assert attempt_meta["adaptive_online_policy_mode"] == "on"
    assert attempt_meta["adaptive_online_action_selected"] == ACTION_SKIP_HELPER
    assert attempt_meta["adaptive_online_action_applied"] == ACTION_SKIP_HELPER
    assert attempt_meta["adaptive_online_updated"] is True
    assert attempt_meta["adaptive_online_update_reason"] == "on_policy_update"
    assert attempt_meta["adaptive_online_event_logged"] is True
    events = [line for line in events_path.read_text(encoding="utf-8").splitlines() if line.strip()]
    assert len(events) == 1


def test_adaptive_stagnation_override_forces_helper_after_policy_skip(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_stagnation_override"
    context_dir = tmp_path / "context_adaptive_stagnation_override"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveGatedCapturePatcher(name="adaptive_stagnation_override")
    controller, _events_path = _skip_biased_online_controller(tmp_path, mode="on")

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    try:
        records = run_eval.run_one_condition(
            case=case,
            condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
            workdir=workdir,
            context_dir_base=context_dir,
            patcher=patcher,
            k=2,
            output_format="json_compact",
            run_id="run-adaptive-stagnation-override",
            include_files_context=False,
            force_patch_attempt=False,
            adaptive_online_policy_controller=controller,
        )
    finally:
        controller.close()

    assert len(records) == 2
    assert patcher.request_calls == 1
    assert records[0]["attempt_metadata"]["adaptive_stagnation_override_applied"] is False
    assert records[1]["attempt_metadata"]["adaptive_stagnation_override_applied"] is True
    assert (
        records[1]["attempt_metadata"]["adaptive_stagnation_override_reason"]
        == "repeated_no_progress_after_stagnation_note"
    )
    assert records[1]["attempt_metadata"]["adaptive_helper_invoked"] is True


def test_adaptive_online_feature_vector_includes_extended_and_lagged_signals(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_online_features"
    context_dir = tmp_path / "context_adaptive_online_features"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveDiscretionInvalidPatcher(name="adaptive_online_feature_keys")

    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    records = run_eval.run_one_condition(
        case=case,
        condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
        workdir=workdir,
        context_dir_base=context_dir,
        patcher=patcher,
        k=2,
        output_format="json_compact",
        run_id="run-adaptive-online-feature-keys",
        include_files_context=False,
        force_patch_attempt=False,
        adaptive_informativeness_pregate=True,
    )

    assert len(records) == 2
    feature_vector_1 = records[0]["attempt_metadata"]["adaptive_online_feature_vector"]
    feature_vector_2 = records[1]["attempt_metadata"]["adaptive_online_feature_vector"]
    expected_keys = {
        "attempts_remaining_frac",
        "helper_budget_remaining_frac",
        "helper_topup_used_frac",
        "helper_topup_remaining_frac",
        "context_compaction_enabled",
        "supports_context_compaction",
        "context_compaction_calls_used_frac",
        "context_compaction_calls_remaining_frac",
        "last_context_compaction_requested",
        "last_context_compaction_invoked",
        "last_context_compaction_success",
        "last_context_compaction_ratio",
        "last_helper_need_more_context",
        "last_helper_invalid_response",
        "last_helper_declined",
        "case_difficulty_ord",
        "case_snapshot_dependency_ord",
        "case_bug_source_imported",
        "pregate_exception_is_assertion",
        "pregate_exception_is_timeout",
        "pregate_exception_is_syntax",
    }
    assert expected_keys.issubset(feature_vector_1.keys())
    assert expected_keys.issubset(feature_vector_2.keys())
    assert feature_vector_1["attempts_remaining_frac"] == 0.5
    assert feature_vector_2["attempts_remaining_frac"] == 0.0
    assert feature_vector_1["case_difficulty_ord"] == 0.0
    assert feature_vector_1["case_snapshot_dependency_ord"] == 0.5
    assert feature_vector_1["case_bug_source_imported"] == 0.0
    assert feature_vector_1["pregate_exception_is_assertion"] == 1.0
    assert feature_vector_1["last_helper_invalid_response"] == 0.0
    assert feature_vector_2["last_helper_invalid_response"] == 1.0


def test_adaptive_online_policy_logging_failure_does_not_abort_attempt(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    workdir = tmp_path / "workdir_adaptive_online_logging_failure"
    context_dir = tmp_path / "context_adaptive_online_logging_failure"
    shutil.copytree(case.path, workdir)
    patcher = _AdaptiveGatedCapturePatcher(name="adaptive_online_logging_failure")
    controller, _events_path = _skip_biased_online_controller(tmp_path, mode="on")

    def _raise_record_outcome(**kwargs):
        del kwargs
        raise OSError("simulated filesystem failure")

    monkeypatch.setattr(controller, "record_outcome", _raise_record_outcome)
    monkeypatch.setattr(
        run_eval,
        "run_pytest",
        lambda *args, **kwargs: run_eval.RunResult(
            returncode=1,
            duration_sec=0.01,
            stdout="tests/test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
            stderr="",
        ),
    )
    monkeypatch.setattr(
        run_eval, "read_latest_snapshot", lambda *_args, **_kwargs: (_snapshot_payload(), None)
    )

    try:
        records = run_eval.run_one_condition(
            case=case,
            condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
            workdir=workdir,
            context_dir_base=context_dir,
            patcher=patcher,
            k=1,
            output_format="json_compact",
            run_id="run-adaptive-online-logging-failure",
            include_files_context=False,
            force_patch_attempt=False,
            adaptive_online_policy_controller=controller,
        )
    finally:
        controller.close()

    assert len(records) == 1
    attempt_meta = records[0]["attempt_metadata"]
    assert attempt_meta["adaptive_online_updated"] is False
    assert attempt_meta["adaptive_online_event_logged"] is False
    assert attempt_meta["adaptive_online_update_reason"] == "event_logging_error:OSError"



from __future__ import annotations

import threading
from collections import defaultdict, deque
from dataclasses import dataclass
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import Any

from llmdebug import _json as json


@dataclass(frozen=True)
class StubbedResponse:
    status: int
    body: dict[str, Any] | list[Any] | str | None
    headers: dict[str, str]


class OpenAIStubServer:
    def __init__(self) -> None:
        self._responses: dict[tuple[str, str], deque[StubbedResponse]] = defaultdict(deque)
        self._requests: list[dict[str, Any]] = []
        self._lock = threading.Lock()
        self._server: ThreadingHTTPServer | None = None
        self._thread: threading.Thread | None = None

    def enqueue_json(
        self,
        *,
        method: str,
        path: str,
        body: dict[str, Any] | list[Any],
        status: int = 200,
        headers: dict[str, str] | None = None,
    ) -> None:
        response_headers = dict(headers or {})
        response_headers.setdefault("Content-Type", "application/json")
        key = (method.upper(), path)
        self._responses[key].append(
            StubbedResponse(status=int(status), body=body, headers=response_headers)
        )

    def enqueue_text(
        self,
        *,
        method: str,
        path: str,
        body: str,
        status: int = 200,
        headers: dict[str, str] | None = None,
    ) -> None:
        response_headers = dict(headers or {})
        response_headers.setdefault("Content-Type", "text/plain; charset=utf-8")
        key = (method.upper(), path)
        self._responses[key].append(
            StubbedResponse(status=int(status), body=body, headers=response_headers)
        )

    @property
    def base_url(self) -> str:
        if self._server is None:
            raise RuntimeError("stub server is not running")
        host, port = self._server.server_address
        return f"http://{host}:{port}"

    @property
    def base_url_v1(self) -> str:
        return f"{self.base_url}/v1"

    @property
    def requests(self) -> list[dict[str, Any]]:
        with self._lock:
            return list(self._requests)

    def requests_for(self, *, method: str, path: str) -> list[dict[str, Any]]:
        method_upper = method.upper()
        with self._lock:
            return [
                dict(item)
                for item in self._requests
                if str(item.get("method")) == method_upper and str(item.get("path")) == path
            ]

    def last_request(self, *, method: str, path: str) -> dict[str, Any] | None:
        matches = self.requests_for(method=method, path=path)
        if not matches:
            return None
        return matches[-1]

    def _handler_cls(self):
        outer = self

        class _Handler(BaseHTTPRequestHandler):
            def do_GET(self) -> None:
                self._handle()

            def do_POST(self) -> None:
                self._handle()

            def log_message(self, format: str, *args: Any) -> None:
                del format, args
                return

            def _handle(self) -> None:
                content_length_raw = self.headers.get("Content-Length")
                try:
                    content_length = int(content_length_raw or 0)
                except ValueError:
                    content_length = 0
                body_bytes = self.rfile.read(content_length) if content_length > 0 else b""
                body_text = body_bytes.decode("utf-8", errors="replace")
                body_json: dict[str, Any] | list[Any] | None = None
                if body_text:
                    try:
                        parsed = json.loads(body_text)
                    except Exception:
                        parsed = None
                    if isinstance(parsed, (dict, list)):
                        body_json = parsed

                request_record: dict[str, Any] = {
                    "method": self.command.upper(),
                    "path": self.path,
                    "headers": {str(k).lower(): str(v) for k, v in self.headers.items()},
                    "body": body_text,
                    "json": body_json,
                }
                with outer._lock:
                    outer._requests.append(request_record)
                    queue = outer._responses[(self.command.upper(), self.path)]
                    if queue:
                        response = queue.popleft()
                    else:
                        response = StubbedResponse(
                            status=404,
                            body={
                                "error": "no_stubbed_response",
                                "method": self.command.upper(),
                                "path": self.path,
                            },
                            headers={"Content-Type": "application/json"},
                        )

                response_body = response.body
                if isinstance(response_body, str):
                    payload = response_body.encode("utf-8")
                elif response_body is None:
                    payload = b""
                else:
                    payload = json.dumps(response_body).encode("utf-8")

                self.send_response(int(response.status))
                for key, value in response.headers.items():
                    self.send_header(str(key), str(value))
                self.send_header("Content-Length", str(len(payload)))
                self.end_headers()
                if payload:
                    self.wfile.write(payload)

        return _Handler

    def __enter__(self) -> OpenAIStubServer:
        if self._server is not None:
            raise RuntimeError("stub server already running")
        self._server = ThreadingHTTPServer(("127.0.0.1", 0), self._handler_cls())
        self._thread = threading.Thread(target=self._server.serve_forever, daemon=True)
        self._thread.start()
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        del exc_type, exc, tb
        server = self._server
        thread = self._thread
        self._server = None
        self._thread = None
        if server is None:
            return
        server.shutdown()
        server.server_close()
        if thread is not None:
            thread.join(timeout=5.0)



"""Additional branch-heavy tests for serialize helpers."""

from __future__ import annotations

import types
from typing import Any

import numpy as np
import pytest

from llmdebug.config import SnapshotConfig
from llmdebug.serialize import (
    _build_local_metadata_entry,
    _compute_array_stats_and_anomalies,
    _jax_stats_and_anomalies,
    _numpy_like_stats_and_anomalies,
    _sample_array_head,
    _torch_stats_and_anomalies,
    get_array_element_count,
    get_requires_grad,
    get_tensor_device,
    serialize_locals_bundle,
    summarize_dataframe,
    summarize_series,
    to_jsonlike,
)


def test_get_tensor_device_paths() -> None:
    class _CpuDevice:
        type = "cpu"

        def __str__(self) -> str:
            return "cpu"

    class _CudaDevice:
        type = "cuda"

        def __str__(self) -> str:
            return "cuda:0"

    class _WithCpuDevice:
        device = _CpuDevice()

    class _WithCudaDevice:
        device = _CudaDevice()

    class _WithDevices:
        def devices(self) -> set[str]:
            return {"gpu:0"}

    class _WithBadDevice:
        @property
        def device(self) -> object:
            raise RuntimeError("boom")

    assert get_tensor_device(_WithCpuDevice()) is None
    assert get_tensor_device(_WithCudaDevice()) == "cuda:0"
    assert get_tensor_device(_WithDevices()) == "{'gpu:0'}"
    assert get_tensor_device(_WithBadDevice()) is None


def test_get_requires_grad_paths() -> None:
    class _TrueGrad:
        requires_grad = True

    class _FalseGrad:
        requires_grad = False

    class _BadGrad:
        @property
        def requires_grad(self) -> bool:
            raise RuntimeError("boom")

    assert get_requires_grad(_TrueGrad()) is True
    assert get_requires_grad(_FalseGrad()) is None
    assert get_requires_grad(_BadGrad()) is None


def test_get_array_element_count_fallbacks() -> None:
    class _FromShape:
        size = "not-an-int"
        shape = (3, 4)

    class _NegativeShape:
        shape = (3, -1)

    class _UnknownShape:
        shape = ("?", 4)

    assert get_array_element_count(_FromShape()) == 12
    assert get_array_element_count(_NegativeShape()) is None
    assert get_array_element_count(_UnknownShape()) is None
    assert get_array_element_count(object()) is None


def test_sample_array_head_fallback_and_error_paths() -> None:
    class _RavelOnly:
        def reshape(self, _shape: int) -> np.ndarray:
            raise RuntimeError("no reshape")

        def ravel(self) -> np.ndarray:
            return np.array([1, 2, 3], dtype=np.int32)

    class _FlattenOnly:
        def flatten(self) -> list[int]:
            return [4, 5, 6]

    assert _sample_array_head(_RavelOnly(), max_items=2) == [1.0, 2.0]
    assert _sample_array_head(_FlattenOnly(), max_items=2) == [4.0, 5.0]

    with pytest.raises(TypeError, match="array does not support"):
        _sample_array_head(object(), max_items=2)


def test_compute_array_stats_and_anomalies_short_circuit_paths() -> None:
    arr = np.arange(10, dtype=np.float32)
    assert _compute_array_stats_and_anomalies(arr, max_elements=3) == (None, None)
    assert _compute_array_stats_and_anomalies(np.array([], dtype=np.float32)) == (None, None)

    class _Unknown:
        __module__ = "custom.module"

    assert _compute_array_stats_and_anomalies(_Unknown()) == (None, None)


def test_torch_stats_and_anomalies_with_stub(monkeypatch: pytest.MonkeyPatch) -> None:
    class _Scalar:
        def __init__(self, value: float) -> None:
            self._value = value

        def item(self) -> float:
            return self._value

    class _Tensor:
        def __init__(self, values: list[float]) -> None:
            self.values = np.asarray(values, dtype=np.float32)

        def detach(self) -> _Tensor:
            return self

        def cpu(self) -> _Tensor:
            return self

        def float(self) -> _Tensor:
            return self

    class _TensorMask:
        def __init__(self, mask: np.ndarray) -> None:
            self.mask = mask

        def sum(self) -> _Scalar:
            return _Scalar(float(np.sum(self.mask)))

    fake_torch = types.SimpleNamespace(
        isnan=lambda t: _TensorMask(np.isnan(t.values)),
        isinf=lambda t: _TensorMask(np.isinf(t.values)),
        min=lambda t: _Scalar(float(np.min(t.values))),
        max=lambda t: _Scalar(float(np.max(t.values))),
        mean=lambda t: _Scalar(float(np.mean(t.values))),
        std=lambda t, unbiased=False: _Scalar(float(np.std(t.values))),
    )
    monkeypatch.setitem(__import__("sys").modules, "torch", fake_torch)

    stats, anomalies = _torch_stats_and_anomalies(_Tensor([1.0, float("nan"), float("inf")]), is_numeric_dtype=True)
    assert stats is not None
    assert np.isnan(stats["min"])
    assert anomalies == {"nan": 1, "inf": 1}

    stats2, anomalies2 = _torch_stats_and_anomalies(_Tensor([1.0]), is_numeric_dtype=False)
    assert stats2 is None
    assert anomalies2 is None


def test_jax_stats_and_anomalies_with_stub(monkeypatch: pytest.MonkeyPatch) -> None:
    fake_jnp = types.SimpleNamespace(
        float32=np.float32,
        isnan=np.isnan,
        isinf=np.isinf,
        min=np.min,
        max=np.max,
        mean=np.mean,
        std=np.std,
    )
    fake_jax = types.SimpleNamespace(numpy=fake_jnp)
    monkeypatch.setitem(__import__("sys").modules, "jax", fake_jax)
    monkeypatch.setitem(__import__("sys").modules, "jax.numpy", fake_jnp)

    class _JaxArray:
        def __init__(self, values: list[float]) -> None:
            self._values = np.asarray(values, dtype=np.float32)

        def astype(self, dtype: Any) -> np.ndarray:
            return self._values.astype(dtype)

    stats, anomalies = _jax_stats_and_anomalies(_JaxArray([1.0, float("nan"), float("inf")]), is_numeric_dtype=True)
    assert stats is not None
    assert anomalies == {"nan": 1, "inf": 1}

    stats2, anomalies2 = _jax_stats_and_anomalies(_JaxArray([1.0]), is_numeric_dtype=False)
    assert stats2 is None
    assert anomalies2 is None


def test_numpy_like_stats_and_anomalies_cupy_path() -> None:
    class _FakeCupy:
        __module__ = "cupy.core.core"

        def __init__(self, values: list[float]) -> None:
            self._values = np.asarray(values, dtype=np.float32)
            self.dtype = self._values.dtype

        def get(self) -> np.ndarray:
            return self._values

    stats, anomalies = _numpy_like_stats_and_anomalies(
        _FakeCupy([1.0, 2.0, 3.0]),
        module="cupy.core.core",
        is_numeric_dtype=True,
    )
    assert stats is not None
    assert anomalies is None


def test_dataframe_and_series_summaries_with_fakes() -> None:
    class _Values:
        def tolist(self) -> list[list[int]]:
            return [[1, 2], [3, 4]]

    class _Head:
        values = _Values()

    class DataFrame:
        __module__ = "pandas.core.frame"

        def __init__(self) -> None:
            self.shape = (2, 2)
            self.columns = ["a", "b", "c"]

        def __len__(self) -> int:
            return 2

        def head(self, _rows: int) -> _Head:
            return _Head()

    class Series:
        __module__ = "pandas.core.series"

        def __init__(self) -> None:
            self.shape = (6,)
            self.name = "s"
            self.dtype = "int64"

        def head(self, _rows: int) -> Series:
            return self

        def tolist(self) -> list[int]:
            return [1, 2, 3, 4, 5]

        def __len__(self) -> int:
            return 6

    cfg = SnapshotConfig(max_items=2)
    df_summary = summarize_dataframe(DataFrame(), cfg)
    assert df_summary["__dataframe__"] is True
    assert df_summary["columns"] == ["a", "b"]
    assert df_summary["columns_truncated"] is True
    assert "head" in df_summary

    series_summary = summarize_series(Series(), cfg)
    assert series_summary["__series__"] is True
    assert series_summary["name"] == "s"
    assert series_summary["head_truncated"] is True


def test_dataframe_and_series_error_paths() -> None:
    class DataFrame:
        __module__ = "pandas.core.frame"

        @property
        def shape(self) -> tuple[int, int]:
            raise RuntimeError("shape fail")

        @property
        def columns(self) -> list[str]:
            raise RuntimeError("cols fail")

        def __len__(self) -> int:
            return 0

        def head(self, _rows: int) -> object:
            raise RuntimeError("head fail")

    class Series:
        __module__ = "pandas.core.series"

        @property
        def shape(self) -> tuple[int]:
            raise RuntimeError("shape fail")

        name = "bad"
        dtype = "int64"

        def head(self, _rows: int) -> object:
            raise RuntimeError("head fail")

    cfg = SnapshotConfig()
    df_summary = summarize_dataframe(DataFrame(), cfg)
    assert str(df_summary["shape"]).startswith("<error:")
    assert str(df_summary["columns"]).startswith("<error:")
    assert str(df_summary["head"]).startswith("<error:")

    series_summary = summarize_series(Series(), cfg)
    assert str(series_summary["shape"]).startswith("<error:")
    assert str(series_summary["head"]).startswith("<error:")


def test_jsonlike_dict_and_object_truncation_paths() -> None:
    cfg = SnapshotConfig(max_items=2)

    data = {1: "a", 2: "b", 3: "c"}
    out = to_jsonlike(data, cfg)
    assert isinstance(out, dict)
    assert "...[TRUNC]" in out

    class _NonDictDunder:
        @property
        def __dict__(self) -> int:  # type: ignore[override]
            return 123

        def __repr__(self) -> str:
            return "non-dict-repr"

    repr_out = to_jsonlike(_NonDictDunder(), cfg)
    assert repr_out == "non-dict-repr"

    class _BigObject:
        pass

    obj = _BigObject()
    for i in range(30):
        setattr(obj, f"k{i}", i)
    obj_out = to_jsonlike(obj, cfg)
    assert isinstance(obj_out, dict)
    assert "...[TRUNC]" in obj_out


def test_build_local_metadata_entry_handles_len_and_shape_errors() -> None:
    class _BadShape:
        def __iter__(self):
            raise RuntimeError("shape fail")

    class _ArrayLike:
        __module__ = "numpy"

        def __init__(self) -> None:
            self.shape = _BadShape()

        def __len__(self) -> int:
            raise RuntimeError("len fail")

    entry = _build_local_metadata_entry(_ArrayLike())
    assert "len" not in entry
    assert str(entry["shape"]).startswith("<error:")


def test_serialize_locals_bundle_handles_len_error_and_redaction_return_type(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    import llmdebug.serialize as serialize_module

    class _LenErrorMapping(dict[str, Any]):
        def __len__(self) -> int:
            raise RuntimeError("len fail")

    cfg = SnapshotConfig(max_items=2)
    bundle = serialize_locals_bundle(_LenErrorMapping({"a": 1, "b": 2}), cfg, [])
    assert bundle.truncated is False
    assert bundle.truncated_keys == 0

    monkeypatch.setattr(serialize_module, "redact_jsonlike", lambda *_a, **_k: ["not", "dict"])
    with pytest.raises(TypeError, match="expected dict"):
        serialize_locals_bundle({"x": 1}, cfg, [])



"""Strict reproducibility importer for TrickyBugs."""

from __future__ import annotations

import re
import shutil
from collections import Counter
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from evals.category_mapping import infer_snapshot_dependency
from evals.importers.base import count_patch_lines, generate_oracle_patch, validate_generated_case
from evals.importers.native_repro_helpers import (
    SKIP_UNSAFE_DATASET_PATH,
    SKIP_UNSUPPORTED_NATIVE_COMMAND,
    normalize_native_pytest_command,
    normalize_relative_path,
    write_native_case_files,
)
from evals.importers.types import ImportRunSummary

from .. import _json as json

_JSON_SCAN_BYTES_LIMIT = 10_000_000
_PYTHON_LANGS = {"python", "python3", "py"}
_LANG_KEYS = ("language", "lang", "programming_language")
_BUGGY_KEYS = (
    "buggy_file",
    "buggy_path",
    "bug_path",
    "faulty_file",
    "faulty_path",
    "before_file",
    "before_path",
)
_FIXED_KEYS = (
    "fixed_file",
    "fixed_path",
    "fix_file",
    "fix_path",
    "correct_file",
    "correct_path",
    "after_file",
    "after_path",
)
_TEST_COMMAND_KEYS = (
    "test_command",
    "test_cmd",
    "run_test",
    "run_tests",
    "testSteps",
    "test_steps",
    "test",
)
_PID_KEYS = ("pid", "problem_id", "problemId")
_PROGRAM_KEYS = ("program_id", "programId", "id", "uid")


@dataclass(frozen=True)
class TrickyBugsCandidate:
    pid: str
    program_id: str
    buggy_rel_path: str
    fixed_rel_path: str
    test_command: str | None
    origin: str


def _safe_slug(raw: str) -> str:
    slug = re.sub(r"[^a-zA-Z0-9]+", "_", raw.strip()).strip("_").lower()
    return slug or "unknown"


def _first_string(record: dict[str, Any], keys: Iterable[str]) -> str | None:
    for key in keys:
        value = record.get(key)
        if isinstance(value, str) and value.strip():
            return value.strip()
    return None


def _iter_dict_nodes(payload: Any) -> Iterable[dict[str, Any]]:
    if isinstance(payload, dict):
        yield payload
        for value in payload.values():
            yield from _iter_dict_nodes(value)
        return
    if isinstance(payload, list):
        for item in payload:
            yield from _iter_dict_nodes(item)


def _candidate_from_record(
    *,
    record: dict[str, Any],
    origin: str,
) -> TrickyBugsCandidate | None:
    buggy_raw = _first_string(record, _BUGGY_KEYS)
    fixed_raw = _first_string(record, _FIXED_KEYS)
    if buggy_raw is None or fixed_raw is None:
        return None

    language = (_first_string(record, _LANG_KEYS) or "").strip().lower()
    if language and language not in _PYTHON_LANGS:
        return None

    buggy_rel = normalize_relative_path(buggy_raw)
    fixed_rel = normalize_relative_path(fixed_raw)
    if buggy_rel is None or fixed_rel is None:
        return None
    if not buggy_rel.endswith(".py") or not fixed_rel.endswith(".py"):
        return None

    command = _first_string(record, _TEST_COMMAND_KEYS)

    pid = _safe_slug(_first_string(record, _PID_KEYS) or "unknown")
    program_id = _safe_slug(_first_string(record, _PROGRAM_KEYS) or "unknown")
    return TrickyBugsCandidate(
        pid=pid,
        program_id=program_id,
        buggy_rel_path=buggy_rel,
        fixed_rel_path=fixed_rel,
        test_command=command,
        origin=origin,
    )


def _iter_python_candidates(trickybugs_root: Path) -> list[TrickyBugsCandidate]:
    discovered: list[TrickyBugsCandidate] = []
    for json_path in sorted(trickybugs_root.rglob("*.json")):
        try:
            if json_path.stat().st_size > _JSON_SCAN_BYTES_LIMIT:
                continue
            payload = json.loads(json_path.read_text(encoding="utf-8", errors="replace"))
        except OSError:
            continue
        except json.JSONDecodeError:
            continue

        origin = f"json:{json_path.relative_to(trickybugs_root)}"
        for record in _iter_dict_nodes(payload):
            candidate = _candidate_from_record(record=record, origin=origin)
            if candidate is not None:
                discovered.append(candidate)

    deduped: dict[tuple[str, str, str, str, str], TrickyBugsCandidate] = {}
    for candidate in discovered:
        key = (
            candidate.pid,
            candidate.program_id,
            candidate.buggy_rel_path,
            candidate.fixed_rel_path,
            candidate.test_command or "",
        )
        deduped.setdefault(key, candidate)

    return sorted(
        deduped.values(),
        key=lambda c: (
            c.pid,
            c.program_id,
            c.buggy_rel_path,
            c.fixed_rel_path,
            c.test_command or "",
        ),
    )


def _read_relative_file(root: Path, rel_path: str) -> str | None:
    normalized = normalize_relative_path(rel_path)
    if normalized is None:
        return None
    file_path = root / normalized
    if not file_path.exists() or not file_path.is_file():
        return None
    return file_path.read_text(encoding="utf-8", errors="replace")


def import_trickybugs(
    output_dir: Path,
    *,
    trickybugs_root: Path,
    data_root: Path,
    max_cases: int = 0,
    validate: bool = True,
    dry_run: bool = False,
    jobs: int = 1,
) -> ImportRunSummary:
    """Import TrickyBugs using strict native dataset test command mapping."""
    del data_root
    del jobs

    candidates = _iter_python_candidates(trickybugs_root)
    skip_reasons: Counter[str] = Counter()
    imported_case_ids: list[str] = []
    attempted = 0
    output_dir.mkdir(parents=True, exist_ok=True)

    for candidate in candidates:
        if max_cases > 0 and len(imported_case_ids) >= max_cases:
            break

        attempted += 1
        case_id = f"tb_{_safe_slug(candidate.pid)}_{_safe_slug(candidate.program_id)}"
        case_dir = output_dir / case_id
        if case_dir.exists() and not dry_run:
            skip_reasons["case_already_exists"] += 1
            continue

        if not candidate.test_command:
            skip_reasons["missing_native_test_oracle_mapping"] += 1
            continue

        native_command = normalize_native_pytest_command(candidate.test_command)
        if native_command is None:
            skip_reasons[SKIP_UNSUPPORTED_NATIVE_COMMAND] += 1
            continue

        buggy_code = _read_relative_file(trickybugs_root, candidate.buggy_rel_path)
        fixed_code = _read_relative_file(trickybugs_root, candidate.fixed_rel_path)
        if buggy_code is None:
            skip_reasons["missing_buggy_source"] += 1
            continue
        if fixed_code is None:
            skip_reasons["missing_fixed_source"] += 1
            continue
        if not buggy_code.strip() or not fixed_code.strip():
            skip_reasons["empty_source_file"] += 1
            continue

        test_files: dict[str, str] = {}
        missing_test = False
        for test_path in native_command.test_paths:
            normalized_test = normalize_relative_path(test_path)
            if normalized_test is None:
                skip_reasons[SKIP_UNSAFE_DATASET_PATH] += 1
                missing_test = True
                break
            test_content = _read_relative_file(trickybugs_root, normalized_test)
            if test_content is None:
                skip_reasons["native_test_file_not_found"] += 1
                missing_test = True
                break
            test_files[normalized_test] = test_content
        if missing_test:
            continue

        oracle_patch = generate_oracle_patch(
            buggy_code,
            fixed_code,
            filename=candidate.buggy_rel_path,
        )
        if not oracle_patch.strip():
            skip_reasons["empty_oracle_patch"] += 1
            continue

        fix_lines = count_patch_lines(oracle_patch)
        difficulty = "easy" if fix_lines <= 3 else ("medium" if fix_lines <= 10 else "hard")
        category = "logic_error"
        metadata: dict[str, Any] = {
            "description": (
                f"TrickyBugs PID {candidate.pid} program {candidate.program_id} "
                f"({candidate.origin})"
            ),
            "category": category,
            "difficulty": difficulty,
            "snapshot_dependency": infer_snapshot_dependency(category),
            "expected_exception": "AssertionError",
            "tags": ["trickybugs", candidate.pid],
            "python_args": list(native_command.python_args),
            "timeout_sec": 30,
            "bug_source": "trickybugs",
            "contamination_risk": "medium",
            "expected_fix_lines": fix_lines,
            "source_id": f"trickybugs_{_safe_slug(candidate.pid)}_{_safe_slug(candidate.program_id)}",
            "difficulty_rationale": "Estimated from patch line count.",
            "repro_mode": "native_dataset_test",
            "source_test_command": native_command.source_test_command,
            "source_test_selector": native_command.source_test_selector,
            "source_buggy_commit": f"path:{candidate.buggy_rel_path}",
            "source_fixed_commit": f"path:{candidate.fixed_rel_path}",
            "trickybugs_pid": candidate.pid,
            "trickybugs_program_id": candidate.program_id,
        }

        if dry_run:
            imported_case_ids.append(case_id)
            continue

        try:
            write_native_case_files(
                case_dir,
                buggy_files={candidate.buggy_rel_path: buggy_code},
                test_files=test_files,
                oracle_patch=oracle_patch,
                metadata=metadata,
            )
        except ValueError:
            skip_reasons[SKIP_UNSAFE_DATASET_PATH] += 1
            shutil.rmtree(case_dir, ignore_errors=True)
            continue

        if validate:
            ok_case, _ = validate_generated_case(case_dir, schema_only=True)
            if not ok_case:
                skip_reasons["schema_validation_failed"] += 1
                shutil.rmtree(case_dir, ignore_errors=True)
                continue

        imported_case_ids.append(case_id)

    imported = len(imported_case_ids)
    skipped = max(0, attempted - imported)
    notes = [
        "Strict reproducibility mode: native dataset test commands only.",
        "No synthetic marker-test conversion is used.",
        "Candidates without deterministic native test mappings are skipped.",
    ]
    return ImportRunSummary(
        source="trickybugs",
        attempted=attempted,
        imported=imported,
        skipped=skipped,
        skip_reasons=dict(skip_reasons),
        imported_case_ids=imported_case_ids,
        notes=notes,
    )



from __future__ import annotations

from llmdebug.rca_engine import build_rca_attempt_record


def test_build_rca_attempt_record_creates_initial_attempt() -> None:
    record = build_rca_attempt_record(
        previous=None,
        attempt_id="att-1",
        session_id="sess-1",
        action="hypothesize",
        failure_signature="sig-1",
        evidence_refs=["exception:AssertionError"],
        hypothesis="off-by-one",
        experiment_plan="add boundary check test",
        gate_mode="soft",
        exploratory=False,
        now="2026-03-01T12:00:00Z",
    )
    assert record["attempt_id"] == "att-1"
    assert record["session_id"] == "sess-1"
    assert record["failure_signature"] == "sig-1"
    assert record["repeat_count"] == 1
    assert record["state"] == "hypothesized"
    assert record["created_at"] == "2026-03-01T12:00:00Z"
    assert record["updated_at"] == "2026-03-01T12:00:00Z"
    assert record["gate_feedback"]["allowed"] is True
    assert record["gate_feedback"]["mode"] == "soft"
    assert record["next_required"] == ["Write a minimal experiment or patch plan."]
    assert isinstance(record["allowed_actions"], list)
    assert record["allowed_actions"]


def test_build_rca_attempt_record_merges_previous_context_and_repeats() -> None:
    previous = {
        "attempt_id": "att-0",
        "session_id": "sess-1",
        "state": "hypothesized",
        "failure_signature": "sig-1",
        "repeat_count": 1,
        "evidence_refs": ["exception:AssertionError"],
        "hypothesis": "off-by-one",
        "hypothesis_delta": "new",
        "experiment_plan": "add boundary check test",
        "proposed_change_summary": None,
        "verification_result": None,
        "gate_feedback": {},
        "gate_warnings": [],
        "allowed_actions": ["plan", "verify"],
        "next_required": ["verification_result"],
        "exploratory": False,
        "transition_history": [],
        "created_at": "2026-03-01T11:59:00Z",
        "updated_at": "2026-03-01T11:59:00Z",
    }
    record = build_rca_attempt_record(
        previous=previous,
        attempt_id="att-2",
        session_id="sess-1",
        action="plan",
        failure_signature="sig-1",
        evidence_refs=["crash_frame:buggy.py:12:f"],
        gate_mode="soft",
        exploratory=True,
        now="2026-03-01T12:01:00Z",
    )
    assert record["repeat_count"] == 2
    assert record["created_at"] == "2026-03-01T11:59:00Z"
    assert record["updated_at"] == "2026-03-01T12:01:00Z"
    assert record["exploratory"] is True
    assert record["experiment_plan"] == "add boundary check test"
    assert "exception:AssertionError" in record["evidence_refs"]
    assert "crash_frame:buggy.py:12:f" in record["evidence_refs"]
    assert record["transition_history"]


def test_build_rca_attempt_record_strict_gate_retains_previous_state() -> None:
    previous = {
        "attempt_id": "att-0",
        "session_id": "sess-1",
        "state": "captured",
        "failure_signature": "sig-1",
        "repeat_count": 1,
        "evidence_refs": [],
        "hypothesis": "",
        "hypothesis_delta": "unknown",
        "experiment_plan": "",
        "proposed_change_summary": None,
        "verification_result": None,
        "gate_feedback": {},
        "gate_warnings": [],
        "allowed_actions": ["bind_evidence", "capture"],
        "next_required": ["Bind concrete evidence from traceback/frames."],
        "exploratory": False,
        "transition_history": [],
        "created_at": "2026-03-01T11:59:00Z",
        "updated_at": "2026-03-01T11:59:00Z",
    }
    record = build_rca_attempt_record(
        previous=previous,
        attempt_id="att-1",
        session_id="sess-1",
        action="plan",
        failure_signature="sig-1",
        evidence_refs=[],
        hypothesis="",
        experiment_plan="",
        gate_mode="strict",
        exploratory=False,
        now="2026-03-01T12:01:00Z",
    )
    assert record["state"] == "captured"
    assert record["gate_feedback"]["allowed"] is False
    assert set(record["gate_feedback"]["triggered_checks"]) == {
        "evidence_required",
        "hypothesis_required",
    }



from __future__ import annotations

import hashlib

from evals.run_eval import (
    render_prompt,
    render_stage_a_request_prompt,
    render_stage_b_patch_prompt,
)


def _base_evidence(**overrides: object) -> dict:
    base = {
        "case_id": "case_1",
        "condition": "traceback_only",
        "attempt": 2,
        "repro": ["python", "-m", "pytest", "-q"],
        "stdout": "E   AssertionError: expected 2 got 1\n",
        "stderr": "",
        "failure_signature": "AssertionError:abc123",
        "previous_failure_signature": "AssertionError:def456",
        "failure_signature_repeat": False,
        "previous_attempt_diff": {
            "attempt_prev": "attempt_01",
            "attempt_curr": "attempt_02",
            "failure_signature_same": False,
            "evidence_delta": {"new_frames": [], "dropped_frames": [], "message_changed": True},
            "change_delta": {"touched_files": ["buggy.py"], "touched_functions": ["f"]},
            "hypothesis_delta": "unknown",
            "verification_delta": {
                "old_status": "unknown",
                "new_status": "unknown",
                "status_changed": False,
            },
        },
        "file_index": "buggy.py\ntest_buggy.py",
        "files": "",
        "snapshot_available": False,
        "snapshot_error": None,
    }
    base.update(overrides)
    return base


def _sha256(value: str) -> str:
    return hashlib.sha256(value.encode("utf-8")).hexdigest()


def test_render_prompt_checksum_freeze() -> None:
    prompt = render_prompt(evidence=_base_evidence(), snapshot=None)
    assert _sha256(prompt) == "b2e477f10ca51a4d14af295a1bd96f313dfaac22ec5a4405deb86cdbbfca0c38"


def test_render_stage_a_prompt_checksum_freeze() -> None:
    prompt = render_stage_a_request_prompt(evidence=_base_evidence(), snapshot=None)
    assert _sha256(prompt) == "cca95644ea77f4b2f64f97a5d3f2eaf4474722da06c1314c372ac78823769481"


def test_render_stage_b_prompt_checksum_freeze() -> None:
    prompt = render_stage_b_patch_prompt(
        evidence=_base_evidence(),
        snapshot=None,
        stage_a_context="--- helper.py ---\ndef value():\n    return 2\n",
        include_project_files=False,
    )
    assert _sha256(prompt) == "644fea45258a97fb724067d29221feb014f303c4481d5e1b46e40d946d128b41"



"""Pure RCA transition and state-evolution helpers."""

from __future__ import annotations

from typing import Any, Literal, cast

RCAState = Literal[
    "captured",
    "evidence_bound",
    "hypothesized",
    "planned",
    "changed",
    "verified",
    "resolved",
    "stalled",
]
RCAAction = Literal[
    "capture",
    "bind_evidence",
    "hypothesize",
    "plan",
    "change",
    "verify",
    "resolve",
]

DEFAULT_STATE: RCAState = "captured"
_VALID_ACTIONS: set[str] = {
    "capture",
    "bind_evidence",
    "hypothesize",
    "plan",
    "change",
    "verify",
    "resolve",
}
_STATE_TRANSITIONS: dict[str, RCAState] = {
    "capture": "captured",
    "bind_evidence": "evidence_bound",
    "hypothesize": "hypothesized",
    "plan": "planned",
    "change": "changed",
    "verify": "verified",
    "resolve": "resolved",
}


def normalize_action(action: str) -> RCAAction:
    """Normalize and validate an RCA action string."""
    lowered = action.strip().lower()
    if lowered in _VALID_ACTIONS:
        return cast(RCAAction, lowered)
    raise ValueError(
        "Invalid RCA action "
        f"{action!r}. Valid actions: {', '.join(sorted(_VALID_ACTIONS))}"
    )


def classify_hypothesis_delta(previous: str, current: str) -> str:
    """Classify how the current hypothesis changed vs previous."""
    prev = previous.strip()
    curr = current.strip()
    if not prev and not curr:
        return "unknown"
    if prev == curr:
        return "unchanged"
    if prev and curr and prev in curr:
        return "refined"
    if not prev and curr:
        return "added"
    if prev and not curr:
        return "removed"
    return "replaced"


def allowed_actions_for_state(state: str) -> list[str]:
    """Return sorted allowed action list for display/UI guidance."""
    matrix: dict[str, list[str]] = {
        "captured": ["bind_evidence", "capture"],
        "evidence_bound": ["hypothesize", "bind_evidence"],
        "hypothesized": ["plan", "bind_evidence", "hypothesize"],
        "planned": ["change", "plan", "bind_evidence"],
        "changed": ["verify", "change", "bind_evidence"],
        "verified": ["resolve", "bind_evidence", "plan"],
        "resolved": ["capture", "bind_evidence"],
        "stalled": ["bind_evidence", "hypothesize", "plan"],
    }
    return matrix.get(state, ["bind_evidence"])


def default_next_required(state: str) -> list[str]:
    """Return canonical coaching steps for the given state."""
    steps: dict[str, list[str]] = {
        "captured": ["Bind concrete evidence from traceback/frames."],
        "evidence_bound": ["State a falsifiable root-cause hypothesis."],
        "hypothesized": ["Write a minimal experiment or patch plan."],
        "planned": ["Apply one focused change tied to the plan."],
        "changed": ["Run verification checks and capture result."],
        "verified": ["Resolve if checks passed, else return to evidence."],
        "stalled": ["Gather new evidence or revise hypothesis before retry."],
    }
    return steps.get(state, [])


def merge_evidence_refs(
    previous: list[str],
    incoming: list[str],
    *,
    max_items: int = 16,
) -> list[str]:
    """Merge + deduplicate evidence refs while keeping stable order."""
    merged: list[str] = []
    seen: set[str] = set()
    for value in previous + incoming:
        if not value:
            continue
        item = value.strip()
        if not item or item in seen:
            continue
        seen.add(item)
        merged.append(item)
        if len(merged) >= max_items:
            break
    return merged


def resolve_previous_values(
    previous: dict[str, Any] | None,
) -> tuple[str, str, str, Any, Any, str]:
    """Extract previous attempt values with stable defaults."""
    previous_state = str(previous.get("state")) if previous else DEFAULT_STATE
    previous_hypothesis = str(previous.get("hypothesis") or "") if previous else ""
    previous_plan = str(previous.get("experiment_plan") or "") if previous else ""
    previous_refs: list[Any] = (
        cast(list[Any], previous.get("evidence_refs"))
        if isinstance(previous, dict) and previous.get("evidence_refs")
        else []
    )
    previous_verification = (
        previous.get("verification_result") if isinstance(previous, dict) else None
    )
    previous_change_summary = (
        str(previous.get("proposed_change_summary") or "") if isinstance(previous, dict) else ""
    )
    return (
        previous_state,
        previous_hypothesis,
        previous_plan,
        previous_refs,
        previous_verification,
        previous_change_summary,
    )


def resolve_merged_evidence_refs(
    previous_refs: Any,
    evidence_refs: list[str] | None,
) -> list[str]:
    """Merge previously stored evidence refs with incoming refs."""
    incoming_refs = evidence_refs if isinstance(evidence_refs, list) else []
    return merge_evidence_refs(
        [str(ref) for ref in cast(list[Any], previous_refs)]
        if isinstance(previous_refs, list)
        else [],
        [str(ref) for ref in incoming_refs],
    )


def resolve_attempt_values(
    *,
    previous_hypothesis: str,
    previous_plan: str,
    previous_change_summary: str,
    previous_verification: Any,
    hypothesis: str | None,
    experiment_plan: str | None,
    proposed_change_summary: str | None,
    verification_result: dict[str, Any] | None,
) -> tuple[str, str, Any, str]:
    """Resolve effective attempt values (input overrides previous)."""
    hypothesis_value = (hypothesis or "").strip() or previous_hypothesis
    plan_value = (experiment_plan or "").strip() or previous_plan
    verification_value = (
        verification_result if verification_result is not None else previous_verification
    )
    proposed_change_value = (proposed_change_summary or "").strip() or previous_change_summary
    return hypothesis_value, plan_value, verification_value, proposed_change_value


def derive_attempt_progress(
    *,
    previous: dict[str, Any] | None,
    failure_signature: str,
    previous_hypothesis: str,
    hypothesis_value: str,
    previous_plan: str,
    plan_value: str,
) -> tuple[bool, int, str, bool, bool]:
    """Compute repeat counts and progression signals for current attempt."""
    same_signature = (
        bool(previous) and str(previous.get("failure_signature") or "") == failure_signature
    )
    repeat_count = int(previous.get("repeat_count") or 1) + 1 if same_signature and previous else 1
    hypothesis_delta = classify_hypothesis_delta(previous_hypothesis, hypothesis_value)
    plan_advanced = bool(plan_value) and plan_value != previous_plan
    hypothesis_advanced = hypothesis_delta in {"added", "refined", "replaced"}
    return same_signature, repeat_count, hypothesis_delta, plan_advanced, hypothesis_advanced


def resolve_target_state(
    previous_state: str,
    normalized_action: RCAAction,
    gate_allowed: bool,
) -> RCAState:
    """Resolve next state according to transition matrix + gate decision."""
    target_state = _STATE_TRANSITIONS.get(normalized_action, "evidence_bound")
    if gate_allowed:
        return target_state
    return cast(
        RCAState,
        previous_state if previous_state in _STATE_TRANSITIONS.values() else "captured",
    )


def apply_stall_rule(
    *,
    target_state: RCAState,
    gate_warnings: list[str],
    same_signature: bool,
    repeat_count: int,
    stall_threshold: int,
    hypothesis_advanced: bool,
    plan_advanced: bool,
    normalized_action: RCAAction,
) -> tuple[RCAState, list[str]]:
    """Apply stall detection for repeated signatures without progression."""
    if (
        same_signature
        and repeat_count >= max(stall_threshold, 2)
        and not hypothesis_advanced
        and not plan_advanced
        and normalized_action in {"bind_evidence", "change", "verify"}
    ):
        gate_warnings.append(
            "Repeated failure signature without hypothesis/plan advancement; state marked stalled."
        )
        return "stalled", gate_warnings
    return target_state, gate_warnings


def build_transition_history(
    *,
    previous: dict[str, Any] | None,
    now: str,
    normalized_action: RCAAction,
    previous_state: str,
    target_state: RCAState,
    gate_allowed: bool,
    gate_mode: str,
) -> list[dict[str, Any]]:
    """Append one transition row while capping history length."""
    previous_history_raw = (
        previous.get("transition_history") if isinstance(previous, dict) else cast(list[Any], [])
    )
    transition_history: list[dict[str, Any]] = []
    if isinstance(previous_history_raw, list):
        transition_history = [
            cast(dict[str, Any], item)
            for item in cast(list[Any], previous_history_raw)
            if isinstance(item, dict)
        ]
    transition_history.append(
        {
            "at": now,
            "action": normalized_action,
            "from": previous_state,
            "to": target_state,
            "gate_allowed": gate_allowed,
            "gate_mode": gate_mode,
        }
    )
    if len(transition_history) > 32:
        transition_history = transition_history[-32:]
    return transition_history



from __future__ import annotations

from pathlib import Path
from typing import Any

from evals.eval_summary import main
from llmdebug import _json as json


def _record(
    *,
    run_id: str,
    case_id: str,
    condition: str,
    success: bool,
    snapshot_dependency: str = "required",
    category: str = "logic_error",
    difficulty: str = "easy",
    attempt: int = 1,
    schema_version: int = 3,
    tokens_prompt: int = 10,
    tokens_completion: int = 2,
    tokens_total: int = 12,
    llm_call_count: int = 1,
    cumulative_runtime_sec: float = 1.0,
    attempt_metadata_updates: dict[str, Any] | None = None,
    patch_meta: dict[str, Any] | None = None,
) -> dict[str, Any]:
    attempt_metadata: dict[str, Any] = {
        "attempt": attempt,
        "max_attempts": 3,
        "attempt_ordinal_label": f"{attempt}/3",
        "llm_call_count": llm_call_count,
        "tokens_prompt": tokens_prompt,
        "tokens_completion": tokens_completion,
        "tokens_total": tokens_total,
        "retry_history_budget_chars": 12000,
        "retry_history_budget_applied": False,
        "prompt_sha256": "prompt-hash",
        "prompt_path": None,
        "evidence_sha256": "evidence-hash",
        "evidence_path": None,
    }
    if attempt_metadata_updates:
        attempt_metadata.update(attempt_metadata_updates)

    record: dict[str, Any] = {
        "schema_version": schema_version,
        "run_id": run_id,
        "case_id": case_id,
        "condition": condition,
        "attempt": attempt,
        "success": success,
        "category": category,
        "difficulty": difficulty,
        "snapshot_dependency": snapshot_dependency,
        "run_metadata": {
            "run_id": run_id,
            "started_at_utc": "2026-03-01T00:00:00Z",
            "deterministic": False,
            "seed": None,
            "python_version": "3.11.0",
            "platform": "test",
            "llmdebug_version": "2.11.4",
            "patcher_name": "openai",
            "patcher_config_fingerprint": "fp",
        },
        "attempt_metadata": attempt_metadata,
        "model_metadata": {
            "backend": "openai",
            "name": "openai",
            "base_url": "http://127.0.0.1:12000/v1",
            "model": "dummy",
            "temperature": 0.0,
            "max_tokens": 2048,
            "response_mode": "edits",
            "api_compat_seed_requested": None,
            "api_compat_seed_acknowledged": False,
        },
        "cumulative_runtime_sec": cumulative_runtime_sec,
        "note": "success" if success else "tests_still_failing",
    }
    if patch_meta is not None:
        record["patch_meta"] = patch_meta
    return record


def _write_jsonl(path: Path, records: list[dict[str, Any]]) -> None:
    path.write_text(
        "\n".join(json.dumps(record, ensure_ascii=True) for record in records) + "\n",
        encoding="utf-8",
    )


def test_eval_summary_missing_file(tmp_path: Path, capsys) -> None:
    rc = main(["missing-run", "--results-dir", str(tmp_path)])
    assert rc == 1
    out = capsys.readouterr().out
    assert "results file not found" in out


def test_eval_summary_rejects_legacy_schema_by_default(tmp_path: Path, capsys) -> None:
    run_id = "legacy-run"
    _write_jsonl(
        tmp_path / f"{run_id}.jsonl",
        [
            _record(
                run_id=run_id,
                case_id="case-legacy",
                condition="traceback_only",
                success=False,
                schema_version=2,
            )
        ],
    )
    rc = main([run_id, "--results-dir", str(tmp_path)])
    assert rc == 2
    out = capsys.readouterr().out
    assert "unsupported schema_version=2" in out


def test_eval_summary_traceback_only_core_output(tmp_path: Path, capsys) -> None:
    run_id = "tb-only-run"
    _write_jsonl(
        tmp_path / f"{run_id}.jsonl",
        [
            _record(
                run_id=run_id,
                case_id="case-a",
                condition="traceback_only",
                success=True,
                tokens_total=30,
                cumulative_runtime_sec=2.0,
            ),
            _record(
                run_id=run_id,
                case_id="case-b",
                condition="traceback_only",
                success=False,
                tokens_total=45,
                cumulative_runtime_sec=3.0,
            ),
        ],
    )
    rc = main([run_id, "--results-dir", str(tmp_path)])
    assert rc == 0
    out = capsys.readouterr().out
    assert "Run Summary" in out
    assert "Condition Effectiveness & Efficiency" in out
    assert "- traceback_only:" in out
    assert "Pairwise Uplift Matrix" in out
    assert "Reliability & Input Quality" in out
    assert "Controls & Guardrails" in out


def test_eval_summary_prints_pairwise_uplift_matrix(tmp_path: Path, capsys) -> None:
    run_id = "pairwise-run"
    _write_jsonl(
        tmp_path / f"{run_id}.jsonl",
        [
            _record(
                run_id=run_id,
                case_id="case-1",
                condition="traceback_only",
                success=False,
            ),
            _record(
                run_id=run_id,
                case_id="case-1",
                condition="with_snapshot",
                success=True,
            ),
            _record(
                run_id=run_id,
                case_id="case-2",
                condition="traceback_only",
                success=True,
            ),
            _record(
                run_id=run_id,
                case_id="case-2",
                condition="with_snapshot",
                success=True,
            ),
        ],
    )
    rc = main([run_id, "--results-dir", str(tmp_path)])
    assert rc == 0
    out = capsys.readouterr().out
    assert "traceback_only -> with_snapshot: +50.0%" in out


def test_eval_summary_prints_adaptive_signals_when_present(tmp_path: Path, capsys) -> None:
    run_id = "adaptive-run"
    _write_jsonl(
        tmp_path / f"{run_id}.jsonl",
        [
            _record(
                run_id=run_id,
                case_id="case-adaptive",
                condition="adaptive_llm_discretion",
                success=True,
                attempt_metadata_updates={
                    "adaptive_mode": "llm_discretion",
                    "adaptive_helper_invoked": True,
                    "adaptive_helper_need_more_context": True,
                    "adaptive_helper_budget_exhausted": True,
                    "adaptive_online_policy_mode": "on",
                    "adaptive_online_policy_algo": "linucb",
                    "adaptive_online_decision_id": "d1",
                    "adaptive_online_action_selected": "invoke_helper",
                    "adaptive_online_action_applied": "invoke_helper",
                    "adaptive_online_updated": True,
                    "adaptive_online_reward": 0.4,
                    "adaptive_online_propensity": 0.8,
                },
                patch_meta={
                    "patch_verifier_enabled": True,
                    "patch_verifier_source_scope_matched": True,
                },
            )
        ],
    )
    rc = main([run_id, "--results-dir", str(tmp_path)])
    assert rc == 0
    out = capsys.readouterr().out
    assert "Adaptive Signals" in out
    assert "online policy: decisions=1" in out
    assert "patch verifier source-scope match rate: 100.0%" in out



"""Additional edge-case tests for CLI helper branches."""

from __future__ import annotations

import os
from pathlib import Path
from types import SimpleNamespace
from typing import Any

import click
import pytest
from click.testing import CliRunner

from llmdebug import cli


class _FakeConsole:
    def __init__(self) -> None:
        self.items: list[Any] = []

    def print(self, item: Any = "", *args: Any, **kwargs: Any) -> None:
        self.items.append(item)


def test_stdout_is_tty_handles_noncallable_and_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    class _NoIsatty:
        isatty = True

    class _BadIsatty:
        def isatty(self) -> bool:
            raise RuntimeError("boom")

    monkeypatch.setattr(click, "get_text_stream", lambda _name: _NoIsatty())
    assert cli._stdout_is_tty() is False

    monkeypatch.setattr(click, "get_text_stream", lambda _name: _BadIsatty())
    assert cli._stdout_is_tty() is False


def test_truncation_helpers_cover_boundary_conditions() -> None:
    assert cli._truncate_text("abcdef", 0) == ""
    assert cli._truncate_text("abcdef", 3) == "abc"
    assert cli._truncate_text("abcdef", 5) == "ab..."

    assert cli._truncate_tail("abcdef", 0) == ""
    assert cli._truncate_tail("abcdef", 3) == "def"
    assert cli._truncate_tail("abcdef", 5) == "...ef"


def test_get_console_returns_console_when_rich_and_tty(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "HAS_RICH", True)
    monkeypatch.setattr(cli, "_stdout_is_tty", lambda: True)
    monkeypatch.setitem(os.environ, "NO_COLOR", "1")
    assert cli._get_console() is not None


def test_snapshot_loader_wrappers_translate_to_system_exit(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        cli,
        "_load_snapshot_impl",
        lambda *_a, **_k: (_ for _ in ()).throw(FileNotFoundError("missing snapshot")),
    )
    with pytest.raises(SystemExit, match="1"):
        cli._load_snapshot(None, ".llmdebug")

    monkeypatch.setattr(
        cli,
        "_load_snapshot_raw_impl",
        lambda *_a, **_k: (_ for _ in ()).throw(FileNotFoundError("missing raw snapshot")),
    )
    with pytest.raises(SystemExit, match="1"):
        cli._load_snapshot_raw(None, ".llmdebug")

    monkeypatch.setattr(
        cli,
        "_load_snapshot_from_file_impl",
        lambda *_a, **_k: (_ for _ in ()).throw(ImportError("toon missing")),
    )
    with pytest.raises(SystemExit, match="1"):
        cli._load_snapshot_from_file(Path("missing.json"))


def test_print_footer_plaintext_includes_git_and_repro(capsys: pytest.CaptureFixture[str]) -> None:
    snapshot = {
        "git": {"commit": "abc1234", "branch": "main", "dirty": True},
        "pytest": {"repro": ["/usr/bin/python3", "-m", "pytest", "tests/test_x.py::test_y", "-q"]},
    }
    cli._print_footer(snapshot, None)
    out = capsys.readouterr().out
    assert "git: abc1234 main *dirty*" in out
    assert "repro: python -m pytest tests/test_x.py::test_y -q" in out


def test_print_exception_panel_handles_malformed_exception(capsys: pytest.CaptureFixture[str]) -> None:
    cli._print_exception_panel({"exception": "not-a-dict"}, None)
    out = capsys.readouterr().out
    assert "UnknownError" in out


def test_print_footer_handles_malformed_git_and_pytest(capsys: pytest.CaptureFixture[str]) -> None:
    cli._print_footer({"git": "bad", "pytest": "bad"}, None)
    out = capsys.readouterr().out
    assert out == ""


def test_print_frame_rich_with_and_without_source(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "HAS_RICH", True)
    console = _FakeConsole()

    with_source = {
        "file": "app.py",
        "line": 10,
        "function": "fn",
        "source": {"start": 10, "end": 10, "snippet": [{"lineno": 10, "code": "x = 1"}]},
        "locals": {"arr": {"__array__": "numpy.ndarray", "shape": [0, 2]}},
        "locals_meta": {"arr": {"type": "numpy.ndarray"}},
    }
    cli._print_frame_rich(with_source, 0, True, console)

    no_source = {"file": "app.py", "line": 11, "function": "fn2", "code": "return x", "locals": {}}
    cli._print_frame_rich(no_source, 1, False, console)

    assert len(console.items) >= 2


def test_print_frame_console_branch_with_locals_and_fallback(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "HAS_RICH", True)
    console = _FakeConsole()
    frame = {
        "file": "app.py",
        "line": 42,
        "function": "run",
        "source": {"start": 41, "end": 43, "snippet": [{"lineno": 42, "code": "x = 1"}]},
        "locals": {"value": 1, "@py_internal": "skip"},
        "locals_meta": {"value": {"type": "builtins.int"}},
    }
    cli._print_frame(frame, 0, True, console, show_locals=True)

    frame_no_parts = {"file": "app.py", "line": 43, "function": "run", "code": "return x"}
    cli._print_frame(frame_no_parts, 1, False, console, show_locals=True)

    assert len(console.items) >= 2


def test_print_diff_rich_covers_change_kinds(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(cli, "HAS_RICH", True)
    console = _FakeConsole()

    changes = [
        SimpleNamespace(kind="modified", name="a", old_value="1", new_value="2"),
        SimpleNamespace(kind="added", name="b", old_value="", new_value="3"),
        SimpleNamespace(kind="removed", name="c", old_value="4", new_value=""),
        SimpleNamespace(
            kind="type_changed",
            name="d",
            old_value="1",
            new_value="'1'",
            old_type="int",
            new_type="str",
        ),
        SimpleNamespace(kind="shape_changed", name="arr", old_value=[1, 2], new_value=[0, 2]),
    ]
    result = SimpleNamespace(
        old_name="old",
        new_name="new",
        same_test=False,
        exception=SimpleNamespace(
            changed=True,
            old_type="ValueError",
            new_type="TypeError",
            old_message="old message",
            new_message="new message",
            old_category="value_error",
            new_category="type_error",
        ),
        location=SimpleNamespace(changed=True, old_location="a.py:1", new_location="b.py:2"),
        locals_diff=SimpleNamespace(changes=changes, unchanged_count=3),
        git_changed=True,
        old_git={"commit": "aaa"},
        new_git={"commit": "bbb"},
    )
    cli._print_diff_rich(result, console)

    unchanged_result = SimpleNamespace(
        old_name="same",
        new_name="same",
        same_test=True,
        exception=SimpleNamespace(
            changed=False,
            old_type="ValueError",
            new_type="ValueError",
            old_message="same message",
            new_message="same message",
            old_category=None,
            new_category=None,
        ),
        location=SimpleNamespace(changed=False, old_location="a.py:1", new_location="a.py:1"),
        locals_diff=SimpleNamespace(changes=[], unchanged_count=0),
        git_changed=False,
        old_git={},
        new_git={},
    )
    cli._print_diff_rich(unchanged_result, console)

    assert len(console.items) >= 5


def test_perform_deletions_counts_success_and_failures(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    ok_file = tmp_path / "ok.json"
    ok_file.write_text("{}", encoding="utf-8")
    missing = tmp_path / "missing.json"

    deleted, failed = cli._perform_deletions([ok_file, missing])
    assert deleted == 1
    assert failed == 1
    assert "could not delete" in capsys.readouterr().err


def test_list_command_rich_branch(monkeypatch: pytest.MonkeyPatch) -> None:
    fake_console = _FakeConsole()
    monkeypatch.setattr(cli, "_get_console", lambda: fake_console)
    monkeypatch.setattr(
        cli,
        "_find_all_snapshots_impl",
        lambda _out_dir, limit=10: [
            {
                "mtime": 1_700_000_000.0,
                "name": "test-name",
                "exc_type": "ValueError",
                "file": "app.py",
                "line": 42,
            }
        ],
    )

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(cli.main, ["list", "--dir", ".llmdebug"])
    assert result.exit_code == 0
    assert fake_console.items


def test_list_command_handles_malformed_entry_fields(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        cli,
        "_find_all_snapshots_impl",
        lambda _out_dir, limit=10: [
            {"mtime": "nan", "name": "bad-row"},
            "not-a-dict",
        ],
    )

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(cli.main, ["list", "--dir", ".llmdebug"])
    assert result.exit_code == 0
    assert "?" in result.output



from __future__ import annotations

from pathlib import Path

from evals import select_cases
from llmdebug import _json as json


def _write_jsonl(path: Path, rows: list[dict]) -> None:
    path.write_text("".join(json.dumps(row) + "\n" for row in rows), encoding="utf-8")


def test_select_cases_failed_mode_with_duplicate_attempt_last_write_wins(tmp_path, capsys) -> None:
    input_path = tmp_path / "run.jsonl"
    output_path = tmp_path / "failed.txt"
    _write_jsonl(
        input_path,
        [
            # case_a attempt 1 originally failing, then duplicate attempt row marks success.
            {"case_id": "case_a", "condition": "traceback_only", "attempt": 1, "success": False},
            {"case_id": "case_a", "condition": "traceback_only", "attempt": 1, "success": True},
            {"case_id": "case_b", "condition": "traceback_only", "attempt": 1, "success": False},
            {"case_id": "case_c", "condition": "traceback_only", "attempt": 2, "success": True},
            {"case_id": "case_d", "condition": "with_snapshot", "attempt": 3, "success": False},
        ],
    )

    exit_code = select_cases.main(
        [
            "--results-jsonl",
            str(input_path),
            "--condition",
            "traceback_only",
            "--mode",
            "failed",
            "--out",
            str(output_path),
        ]
    )

    assert exit_code == 0
    assert output_path.read_text(encoding="utf-8") == "case_b\n"
    assert "Selected 1 case(s)" in capsys.readouterr().out


def test_select_cases_attempt_ge_mode(tmp_path) -> None:
    input_path = tmp_path / "run.jsonl"
    output_path = tmp_path / "attempt_ge.txt"
    _write_jsonl(
        input_path,
        [
            {"case_id": "case_a", "condition": "traceback_only", "attempt": 1, "success": True},
            {"case_id": "case_b", "condition": "traceback_only", "attempt": 2, "success": True},
            {"case_id": "case_c", "condition": "traceback_only", "attempt": 3, "success": False},
        ],
    )

    exit_code = select_cases.main(
        [
            "--results-jsonl",
            str(input_path),
            "--condition",
            "traceback_only",
            "--mode",
            "attempt-ge",
            "--attempt-threshold",
            "2",
            "--out",
            str(output_path),
        ]
    )

    assert exit_code == 0
    assert output_path.read_text(encoding="utf-8") == "case_b\ncase_c\n"


def test_select_cases_failed_or_attempt_ge_mode(tmp_path) -> None:
    input_path = tmp_path / "run.jsonl"
    output_path = tmp_path / "union.txt"
    _write_jsonl(
        input_path,
        [
            {"case_id": "case_a", "condition": "traceback_only", "attempt": 1, "success": False},
            {"case_id": "case_b", "condition": "traceback_only", "attempt": 2, "success": True},
            {"case_id": "case_c", "condition": "traceback_only", "attempt": 1, "success": True},
        ],
    )

    exit_code = select_cases.main(
        [
            "--results-jsonl",
            str(input_path),
            "--condition",
            "traceback_only",
            "--mode",
            "failed-or-attempt-ge",
            "--attempt-threshold",
            "2",
            "--out",
            str(output_path),
        ]
    )

    assert exit_code == 0
    assert output_path.read_text(encoding="utf-8") == "case_a\ncase_b\n"


def test_select_cases_fails_when_condition_missing(tmp_path, capsys) -> None:
    input_path = tmp_path / "run.jsonl"
    output_path = tmp_path / "missing.txt"
    _write_jsonl(
        input_path,
        [
            {"case_id": "case_a", "condition": "with_snapshot", "attempt": 1, "success": False},
        ],
    )

    exit_code = select_cases.main(
        [
            "--results-jsonl",
            str(input_path),
            "--condition",
            "traceback_only",
            "--out",
            str(output_path),
        ]
    )

    assert exit_code == 2
    assert "No rows found for condition='traceback_only'" in capsys.readouterr().out



"""Tests for snapshot loader/index robustness paths."""

from __future__ import annotations

from pathlib import Path

import llmdebug._snapshot_loader as snapshot_loader


def test_find_all_snapshots_skips_invalid_index_mtime(
    tmp_path: Path,
    monkeypatch,
) -> None:
    good_path = tmp_path / "good.json"
    good_path.write_text("{}", encoding="utf-8")
    bad_path = tmp_path / "bad.json"
    bad_path.write_text("{}", encoding="utf-8")

    monkeypatch.setattr(
        snapshot_loader,
        "list_snapshot_records",
        lambda _out_dir, limit=None: [
            {
                "path": str(good_path),
                "mtime": "1700000000.0",
                "name": "good",
                "exc_type": "ValueError",
                "exc_message": "ok",
                "file": "app.py",
                "line": "12",
            },
            {
                "path": str(bad_path),
                "mtime": "nan",
                "name": "bad",
                "exc_type": "TypeError",
                "exc_message": "bad",
                "file": "bad.py",
                "line": "5",
            },
        ],
    )

    removed: list[str] = []
    monkeypatch.setattr(
        snapshot_loader,
        "remove_snapshot_records",
        lambda _out_dir, paths: removed.extend(paths),
    )

    entries = snapshot_loader.find_all_snapshots(str(tmp_path))
    assert len(entries) == 1
    assert entries[0]["path"] == good_path.resolve()
    assert str(bad_path) in removed



"""Tests for RCA backend adapters."""

from __future__ import annotations

from pathlib import Path

from llmdebug import _json as json
from llmdebug.rca_backend import backend_name, get_rca_history, get_rca_status, persist_rca_record


def _record(
    attempt_id: str,
    *,
    session_id: str = "sess-a",
    state: str = "captured",
    updated_at: str = "2026-02-27T10:00:00Z",
) -> dict[str, object]:
    return {
        "attempt_id": attempt_id,
        "session_id": session_id,
        "updated_at": updated_at,
        "created_at": "2026-02-27T10:00:00Z",
        "failure_signature": "sig-a",
        "state": state,
        "evidence_refs": ["r1"],
    }


def test_backend_name_defaults_and_invalid(monkeypatch) -> None:
    monkeypatch.delenv("LLMDEBUG_RCA_BACKEND", raising=False)
    assert backend_name() == "sqlite"

    monkeypatch.setenv("LLMDEBUG_RCA_BACKEND", "not-real")
    assert backend_name() == "sqlite"


def test_jsonl_roundtrip_status_and_history(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.setenv("LLMDEBUG_RCA_BACKEND", "jsonl")
    monkeypatch.setenv("LLMDEBUG_RCA_MAX_RECORDS", "2")

    out_dir = str(tmp_path)
    persist_rca_record(out_dir=out_dir, record=_record("a1", updated_at="2026-02-27T10:00:00Z"))
    persist_rca_record(out_dir=out_dir, record=_record("a2", updated_at="2026-02-27T10:01:00Z"))
    persist_rca_record(out_dir=out_dir, record=_record("a3", updated_at="2026-02-27T10:02:00Z"))

    status = get_rca_status(out_dir=out_dir, session_id="sess-a")
    assert status is not None
    assert status["attempt_id"] == "a3"

    history = get_rca_history(out_dir=out_dir, session_id="sess-a", limit=10)
    assert [entry["attempt_id"] for entry in history] == ["a3", "a2"]

    store = tmp_path / "rca_attempts.jsonl"
    lines = [line for line in store.read_text(encoding="utf-8").splitlines() if line.strip()]
    assert len(lines) == 2


def test_sqlite_migration_guard_skips_when_rows_exist(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.setenv("LLMDEBUG_RCA_BACKEND", "sqlite")
    out_dir = str(tmp_path)

    persist_rca_record(out_dir=out_dir, record=_record("sql-1"))

    jsonl_path = tmp_path / "rca_attempts.jsonl"
    jsonl_path.write_text(json.dumps(_record("jsonl-1")) + "\n", encoding="utf-8")

    history = get_rca_history(out_dir=out_dir, session_id="sess-a", limit=10)
    assert history
    assert jsonl_path.exists()
    assert not Path(f"{jsonl_path}.bak").exists()


def test_sqlite_migration_filters_invalid_records_and_applies_retention(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.setenv("LLMDEBUG_RCA_BACKEND", "sqlite")
    monkeypatch.setenv("LLMDEBUG_RCA_MAX_RECORDS", "2")
    out_dir = str(tmp_path)
    jsonl_path = tmp_path / "rca_attempts.jsonl"
    rows = [
        _record("a1", updated_at="2026-02-27T10:00:00Z"),
        {**_record("bad", updated_at="2026-02-27T10:00:30Z"), "attempt_id": ""},
        _record("a2", updated_at="2026-02-27T10:01:00Z"),
        _record("a3", updated_at="2026-02-27T10:02:00Z"),
    ]
    jsonl_path.write_text("\n".join(json.dumps(row) for row in rows) + "\n", encoding="utf-8")

    history = get_rca_history(out_dir=out_dir, session_id="sess-a", limit=10)
    assert [entry["attempt_id"] for entry in history] == ["a3", "a2"]
    assert not jsonl_path.exists()
    assert Path(f"{jsonl_path}.bak").exists()



"""RCA input normalization helpers for MCP tools.

This module keeps MCP adapter functions thin by centralizing parsing and
context resolution for manual RCA advancement.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ._snapshot_loader import load_snapshot
from .rca_state import build_failure_signature, derive_session_id


@dataclass(frozen=True)
class RCAAdvanceContext:
    """Normalized inputs for one manual RCA advancement request."""

    snapshot: dict[str, Any] | None
    session_id: str
    failure_signature: str
    evidence_refs: list[str]
    verification_result: dict[str, Any] | None
    hypothesis: str | None
    experiment_plan: str | None
    proposed_change_summary: str | None
    override_reason: str | None


class SnapshotLoadError(RuntimeError):
    """Raised when a requested snapshot cannot be loaded for manual RCA flow."""


class RCAInputError(ValueError):
    """Raised when manual RCA inputs are ambiguous or internally inconsistent."""


def _split_csv_values(raw: str) -> list[str]:
    return [item.strip() for item in raw.split(",") if item.strip()]


def _resolve_snapshot(
    *,
    out_dir: str,
    snapshot_path: str,
    session_id: str,
    failure_signature: str,
) -> dict[str, Any] | None:
    if snapshot_path:
        try:
            return load_snapshot(snapshot_path, out_dir)
        except Exception as exc:  # pragma: no cover - exercised through public API
            raise SnapshotLoadError(f"Error loading snapshot: {type(exc).__name__}: {exc}") from exc

    has_session = bool(session_id.strip())
    has_signature = bool(failure_signature.strip())
    if not has_session and not has_signature:
        try:
            return load_snapshot(None, out_dir)
        except Exception:
            return None
    if has_session != has_signature:
        raise RCAInputError(
            "When snapshot_path is omitted, provide both session_id and failure_signature "
            "together, or leave both empty to derive from the latest snapshot."
        )
    return None


def _verification_result(status: str, checks: str) -> dict[str, Any] | None:
    normalized_status = status.strip()
    if not normalized_status:
        return None
    return {
        "status": normalized_status,
        "checks": _split_csv_values(checks),
    }


def _strip_or_none(raw: str) -> str | None:
    normalized = raw.strip()
    return normalized or None


def resolve_rca_advance_context(
    *,
    out_dir: str,
    snapshot_path: str,
    session_id: str,
    failure_signature: str,
    evidence_refs: str,
    hypothesis: str,
    experiment_plan: str,
    proposed_change_summary: str,
    verification_status: str,
    verification_checks: str,
    override_reason: str,
) -> RCAAdvanceContext:
    """Resolve normalized RCA context from MCP tool string inputs."""
    snapshot = _resolve_snapshot(
        out_dir=out_dir,
        snapshot_path=snapshot_path,
        session_id=session_id,
        failure_signature=failure_signature,
    )
    resolved_session = session_id.strip() or derive_session_id(
        snapshot, out_dir=out_dir, fallback="manual"
    )
    signature = failure_signature.strip() or build_failure_signature(snapshot)
    return RCAAdvanceContext(
        snapshot=snapshot,
        session_id=resolved_session,
        failure_signature=signature,
        evidence_refs=_split_csv_values(evidence_refs),
        verification_result=_verification_result(verification_status, verification_checks),
        hypothesis=_strip_or_none(hypothesis),
        experiment_plan=_strip_or_none(experiment_plan),
        proposed_change_summary=_strip_or_none(proposed_change_summary),
        override_reason=_strip_or_none(override_reason),
    )
